from flatdict import FlatDict
import xml.etree.cElementTree as ET
from openpnm.io import Dict, GenericIO
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class XDMF(GenericIO):
    r"""
    The eXtensible Data Model Format combines XML descriptors with HDF5 data
    storage


    For more information visit the webiste:
    `XDMF.org <http://www.xdmf.org/index.php/Main_Page>`_
    """

    _header = '''<?xml version="1.0" ?>
                 <!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>'''

    @classmethod
    def save(cls, network, phases=[], filename=''):
        r"""
        Saves (transient/steady-state) data from the given objects into the
        specified file.

        Parameters
        ----------
        network : OpenPNM Network Object
            The network containing the desired data

        phases : list of OpenPNM Phase Objects (optional, default is none)
            A list of phase objects whose data are to be included

        Notes
        -----
        This method only saves the data, not any of the pore-scale models or
        other attributes.  To save an actual OpenPNM Project use the
        ``Workspace`` object.

        """
        import h5py
        project, network, phases = cls._parse_args(network=network,
                                                   phases=phases)
        network = network[0]
        # Check if any of the phases has time series
        transient = GenericIO.is_transient(phases=phases)

        if filename == '':
            filename = project.name
        path = cls._parse_filename(filename=filename, ext='xmf')
        # Path is a pathlib object, so slice it up as needed
        fname_xdf = path.name
        d = Dict.to_dict(network, phases=phases, interleave=True,
                         flatten=False, categorize_by=['element', 'data'])
        D = FlatDict(d, delimiter='/')
        # Identify time steps
        t_steps = []
        if transient:
            for key in D.keys():
                if '@' in key:
                    t_steps.append(key.split('@')[1])
        t_steps = list(set(t_steps))
        t_grid = create_grid(Name="TimeSeries", GridType="Collection",
                             CollectionType="Temporal")
        # If steady-state, define '0' time step
        if not transient:
            t_steps.append('0')
        # Setup xdmf file
        root = create_root('Xdmf')
        domain = create_domain()
        # Iterate over time steps present
        for t in range(len(t_steps)):
            # Define the hdf file
            if not transient:
                fname_hdf = path.stem+".hdf"
            else:
                fname_hdf = path.stem+'@'+t_steps[t]+".hdf"
            path_p = path.parent
            f = h5py.File(path_p.joinpath(fname_hdf), "w")
            # Add coordinate and connection information to top of HDF5 file
            f["coordinates"] = network["pore.coords"]
            f["connections"] = network["throat.conns"]
            # geometry coordinates
            row, col = f["coordinates"].shape
            dims = ' '.join((str(row), str(col)))
            hdf_loc = fname_hdf + ":coordinates"
            geo_data = create_data_item(value=hdf_loc, Dimensions=dims,
                                        Format='HDF', DataType="Float")
            geo = create_geometry(GeometryType="XYZ")
            geo.append(geo_data)
            # topolgy connections
            row, col = f["connections"].shape  # col first then row
            dims = ' '.join((str(row), str(col)))
            hdf_loc = fname_hdf + ":connections"
            topo_data = create_data_item(value=hdf_loc, Dimensions=dims,
                                         Format="HDF", NumberType="Int")
            topo = create_topology(TopologyType="Polyline",
                                   NodesPerElement=str(2),
                                   NumberOfElements=str(row))
            topo.append(topo_data)
            # Make HDF5 file with all datasets, and no groups
            for item in D.keys():
                if D[item].dtype == 'O':
                    logger.warning(item + ' has dtype object,'
                                   + ' will not write to file')
                    del D[item]
                elif 'U' in str(D[item][0].dtype):
                    pass
                elif ('@' in item and t_steps[t] == item.split('@')[1]):
                    f.create_dataset(name='/'+item.split('@')[0]+'@t',
                                     shape=D[item].shape,
                                     dtype=D[item].dtype,
                                     data=D[item])
                elif ('@' not in item and t == 0):
                    f.create_dataset(name='/'+item, shape=D[item].shape,
                                     dtype=D[item].dtype, data=D[item])
            # Create a grid
            grid = create_grid(Name=t_steps[t], GridType="Uniform")
            time = create_time(type='Single', Value=t_steps[t])
            grid.append(time)
            # Add pore and throat properties
            for item in D.keys():
                if item not in ['coordinates', 'connections']:
                    if ("@" in item and t_steps[t] == item.split("@")[1]) or (
                        "@" not in item
                    ):
                        attr_type = 'Scalar'
                        shape = D[item].shape
                        dims = (''.join([str(i) + ' ' for i in list(shape)[::-1]]))
                        if '@' in item:
                            item = item.split('@')[0]+'@t'
                            hdf_loc = fname_hdf + ":" + item
                        elif ('@' not in item and t == 0):
                            hdf_loc = fname_hdf + ":" + item
                        elif ('@' not in item and t > 0):
                            hdf_loc = path.stem + '@' + t_steps[0] + ".hdf" + ":" + item
                        attr = create_data_item(value=hdf_loc,
                                                Dimensions=dims,
                                                Format='HDF',
                                                Precision='8',
                                                DataType='Float')
                        name = item.replace('/', ' | ')
                        if 'throat' in item:
                            Center = "Cell"
                        else:
                            Center = "Node"
                        el_attr = create_attribute(Name=name, Center=Center,
                                                   AttributeType=attr_type)
                        el_attr.append(attr)
                        grid.append(el_attr)
                    else:
                        pass
            grid.append(topo)
            grid.append(geo)
            t_grid.append(grid)
            # CLose the HDF5 file
            f.close()
        domain.append(t_grid)
        root.append(domain)
        with open(path_p.joinpath(fname_xdf), 'w') as file:
            file.write(cls._header)
            file.write(ET.tostring(root).decode("utf-8"))


def create_root(Name):
    return ET.Element(Name)


def create_domain():
    return ET.Element("Domain")


def create_geometry(GeometryType, **attribs):
    element = ET.Element('Geometry')
    element.attrib.update({'GeometryType': GeometryType})
    element.attrib.update(attribs)
    return element


def create_topology(TopologyType, NumberOfElements, **attribs):
    element = ET.Element('Topology')
    element.attrib.update({'TopologyType': TopologyType,
                           'NumberOfElements': NumberOfElements})
    if TopologyType in ['Polyline']:
        if 'NodesPerElement' not in attribs.keys():
            raise Exception('NodesPerElement must be specified')
    element.attrib.update(attribs)
    return element


def create_attribute(Name, **attribs):
    element = ET.Element('Attribute')
    element.attrib.update({'Name': Name})
    element.attrib.update(attribs)
    return element


def create_time(type='Single', Value=None):
    element = ET.Element('Time')
    if type == 'Single' and Value:
        element.attrib['Value'] = Value
    return element


def create_grid(Name, GridType, **attribs):
    element = ET.Element('Grid')
    element.attrib.update({'Name': Name,
                           'GridType': GridType,
                           'Section': None})
    element.attrib.update(attribs)
    if element.attrib['GridType'] != 'Subset':
        if 'Section' in element.attrib.keys():
            del element.attrib['Section']
    return element


def create_data_item(value, Dimensions, **attribs):
    element = ET.Element('DataItem')
    element.attrib.update({'ItemType': "Uniform",
                           'Format': "XML",
                           'DatarType': "Float",
                           'Precision': "4",
                           'Rank': "1",
                           'Dimensions': Dimensions,
                           'Function': None})
    element.attrib.update(attribs)
    if element.attrib['Function'] is None:
        del element.attrib['Function']
    element.text = value
    return element



from openpnm.phases.mixtures import GenericMixture, species
from openpnm import models as mods
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class SalineWater(GenericMixture):

    def __init__(self, network, components=None, **kwargs):
        if components is not None:
            logger.warn('Ignoring received components')
        super().__init__(network=network, components=[], **kwargs)

        Cl = species.ions.Cl(network=network, name='Cl_'+self.name)
        Na = species.ions.Na(network=network, name='Na_'+self.name)
        W = species.liquids.H2O(network=network, name='H2O_'+self.name)
        self.set_component([Cl, Na, W])
        self.set_concentration(component=W, values=998/0.018)
        self.set_concentration(component=Na, values=0.0)
        self.set_concentration(component=Cl, values=0.0)
        self.update_mole_fractions()
        self.add_model(propname='pore.salt_concentration',
                       model=mods.misc.summation,
                       props=['pore.concentration.Na_'+self.name,
                              'pore.concentration.Cl_'+self.name])
        self.add_model(propname='pore.salinity',
                       model=mods.phases.mixtures.salinity,
                       concentration='pore.concentration.Na_'+self.name)
        self.add_model(propname='pore.mass_density',
                       model=mods.phases.density.water,
                       salinity='pore.salinity')
        self.add_model(propname='pore.viscosity',
                       model=mods.phases.viscosity.water,
                       salinity='pore.salinity')
        self.add_model(propname='pore.permittivity',
                       model=mods.phases.permittivity.water)



from openpnm.algorithms import TransientReactiveTransport, IonicConduction
from openpnm.utils import logging, Docorator, GenericSettings
logger = logging.getLogger(__name__)
docstr = Docorator()


@docstr.get_sectionsf('TransientIonicConductionSettings',
                      sections=['Parameters'])
@docstr.dedent
class TransientIonicConductionSettings(GenericSettings):
    r"""

    Parameters
    ----------
    ##

    Other Parameters
    ----------------

    **The following parameters pertain to steady-state IonicConduction**

    %(IonicConductionSettings.parameters)s

    ----

    **The following parameters pertain to the ReactiveTransport class**

    %(ReactiveTransportSettings.other_parameters)s

    ----

    **The following parameters pertain to the GenericTransport class**

    %(GenericTransportSettings.other_parameters)s

    """
    quantity = 'pore.potential'
    conductance = 'throat.ionic_conductance'
    charge_conservation = 'electroneutrality'
    cache_A = False
    cache_b = False


class TransientIonicConduction(TransientReactiveTransport,
                               IonicConduction):
    r"""
    A subclass of GenericTransport to perform steady and transient simulations
    of pure diffusion and advection-diffusion problems.

    """

    def __init__(self, settings={}, phase=None, **kwargs):
        super().__init__(**kwargs)
        c = TransientIonicConductionSettings()
        self.settings._update_settings_and_docs(c)
        self.settings.update(settings)
        if phase is not None:
            self.setup(phase=phase)

    def setup(self, phase=None, quantity='', charge_conservation=None,
              conductance='', t_initial=None, t_final=None, t_step=None,
              t_output=None, t_tolerance=None, t_precision=None, t_scheme='',
              **kwargs):
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        if charge_conservation:
            self.settings['charge_conservation'] = charge_conservation
        if t_initial is not None:
            self.settings['t_initial'] = t_initial
        if t_final is not None:
            self.settings['t_final'] = t_final
        if t_step is not None:
            self.settings['t_step'] = t_step
        if t_output is not None:
            self.settings['t_output'] = t_output
        if t_tolerance is not None:
            self.settings['t_tolerance'] = t_tolerance
        if t_precision is not None:
            self.settings['t_tolerance'] = t_precision
        if t_scheme:
            self.settings['t_scheme'] = t_scheme
        self.settings.update(kwargs)



r"""

**openpnm.io**

----

This module contains functionality for importing and exporting data between
OpenPNM and other formats.

Note that saving and loading OpenPNM data in its own format is done through
the Workspace and Project objects.  Saving data through these classes losses
lots of vital information about the simulation, such as pore scale-models and
parameters.

----

**Available Import and Export Options**

+----------+------------------------------------------------------------------+
| Format   | Description                                                      |
+==========+==================================================================+
| Dict     | Generates hierarchical ``dicts`` with a high degree of control   |
|          | over the structure                                               |
+----------+------------------------------------------------------------------+
| Pandas   | Combines all data arrays into a Pandas DataFrame object          |
+----------+------------------------------------------------------------------+
| OpenpnmIO| Reads and write the normal OpenPNM formats                       |
+----------+------------------------------------------------------------------+
| CSV      | Reads and writes CSV (comma-separated-value files) containing    |
|          | pore and throat data                                             |
+----------+------------------------------------------------------------------+
| HDF5     | The HDF5 (Hierarchical Data Format) file is good for high-       |
|          | peformance, long term data storage                               |
+----------+------------------------------------------------------------------+
| XDMF     | The eXtensible Data Model Format combines XML descriptors with   |
|          | HDF5 data storage                                                |
+----------+------------------------------------------------------------------+
| VTK      | The Visualization Toolkit (VTK) format defined by Kitware and    |
|          | used by Paraview                                                 |
+----------+------------------------------------------------------------------+
| NetworkX | NetworkX is a common tool for dealing with network structures    |
+----------+------------------------------------------------------------------+
| MAT      | MAT files are a format used by Matlab                            |
+----------+------------------------------------------------------------------+
| JGF      | The JSON Graph Format is a schema specification of how to write  |
|          | a graph object into JSON format                                  |
+----------+------------------------------------------------------------------+
| STL      | The STL Format is a Standard Triangle (or Tessellation) Language |
|          | supported by many CAD packages and used for 3D printing          |
+----------+------------------------------------------------------------------+
| Salome   | A *.py instruction file to be read by Salome to build a geomertry|
+----------+------------------------------------------------------------------+
| PerGeos  | The PerGeos format is used by the commercial software Avizo      |
+----------+------------------------------------------------------------------+
| PoreSpy  | PoreSpy contains the snow extraction algorithm                   |
+----------+------------------------------------------------------------------+
| iMorph   | iMorph is a graphical interface program that provides some image |
|          | analysis tools for porous media                                  |
+----------+------------------------------------------------------------------+
| MARock   | 3DMA-Rock is a network extraction algorithm                      |
+----------+------------------------------------------------------------------+
| Statoil  | The StatOil format is used by the Maximal Ball network extraction|
|          | code of the Imperial College London group                        |
+----------+------------------------------------------------------------------+

Of the above, the last three only support *reading* from files and there are
no plans to implement writing such files.

"""
from .GenericIO import GenericIO
from .OpenpnmIO import OpenpnmIO
from .Dict import Dict
from .VTK import VTK
from .CSV import CSV
from .NetworkX import NetworkX
from .MAT import MAT
from .iMorph import iMorph
from .MARock import MARock
from .Statoil import Statoil
from .PerGeos import PerGeos
from .PoreSpy import PoreSpy
from .Pandas import Pandas
from .HDF5 import HDF5
from .XDMF import XDMF
from .JSONGraphFormat import JSONGraphFormat
from .STL import STL
from .COMSOL import COMSOL
from .Salome import Salome



r"""

.. autofunction:: openpnm.models.phases.thermal_conductivity.water
.. autofunction:: openpnm.models.phases.thermal_conductivity.chung
.. autofunction:: openpnm.models.phases.thermal_conductivity.sato

"""
import numpy as np
import scipy as sp


def water(target, temperature="pore.temperature", salinity="pore.salinity"):
    r"""
    Calculates thermal conductivity of pure water or seawater at atmospheric
    pressure using the correlation given by Jamieson and Tudhope. Values at
    temperature higher  the normal boiling temperature are calculated at the
    saturation pressure.

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    temperature : string
        The dictionary key containing the temperature values.  Temperature must
        be in Kelvin for this emperical equation to work

    salinity : string
        The dictionary key containing the salinity values.  Salinity must be
        expressed in g of salt per kg of solution (ppt).

    Returns
    -------
    value : NumPy ndarray
        Array containing thermal conductivity of water/seawater in [W/m.K]

    Notes
    -----
    T must be in K, and S in g of salt per kg of phase, or ppt (parts per
    thousand)

    VALIDITY: 273 < T < 453 K; 0 < S < 160 g/kg;
    ACCURACY: 3 %

    References
    ----------
    D. T. Jamieson, and J. S. Tudhope, Desalination, 8, 393-401, 1970.

    """
    T = target[temperature]
    if salinity in target.keys():
        S = target[salinity]
    else:
        S = 0
    T68 = 1.00024 * T  # convert from T_90 to T_68
    SP = S / 1.00472  # convert from S to S_P
    k_sw = 0.001 * (
        10
        ** (
            np.log10(240 + 0.0002 * SP)
            + 0.434
            * (2.3 - (343.5 + 0.037 * SP) / T68)
            * ((1 - T68 / (647.3 + 0.03 * SP))) ** (1 / 3)
        )
    )
    value = k_sw
    return value


def chung(
    target,
    Cv="pore.heat_capacity",
    acentric_factor="pore.acentric_factor",
    mol_weight="pore.molecular_weight",
    viscosity="pore.viscosity",
    temperature="pore.temperature",
    critical_temperature="pore.critical_temperature",
):
    r"""
    Uses Chung et al. model to estimate thermal conductivity for gases with
    low pressure(<10 bar) from first principles at conditions of interest

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    acentric_factor : string
        Dictionary key containing the acentric factor of the component

    Cv : string
        Dictionary key containing the heat capacity at constant volume
        (J/(mol.K))

    mol_weight : string
        Dictionary key containing the molecular weight of the component
        (kg/mol)

    viscosity : string
        The dictionary key containing the viscosity values (Pa.s)

    temperature : string
        The dictionary key containing the temperature values (K)

    critical_temperatre: string
        The dictionary key containing the critical temperature values (K)

    Returns
    -------
    value : NumPy ndarray
        Array containing thermal conductivity values in [W/m.K]

    """
    Cv = target[Cv]
    acentric = target[acentric_factor]
    MW = target[mol_weight]
    R = 8.314
    T = target[temperature]
    mu = target[viscosity]
    Tc = target[critical_temperature]
    Tr = T / Tc
    z = 2.0 + 10.5 * Tr ** 2
    beta = 0.7862 - 0.7109 * acentric + 1.3168 * acentric ** 2
    alpha = Cv / R - 3 / 2
    s = 1 + alpha * (
        (0.215 + 0.28288 * alpha - 1.061 * beta + 0.26665 * z)
        / (0.6366 + beta * z + 1.061 * alpha * beta)
    )
    value = 3.75 * s * (mu) * R / (MW)
    return value


def sato(
    target,
    mol_weight="pore.molecular_weight",
    boiling_temperature="pore.boiling_point",
    temperature="pore.temperature",
    critical_temperature="pore.critical_temperature",
):
    r"""
    Uses Sato et al. model to estimate thermal conductivity for pure liquids
    from first principles at conditions of interest

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    boiling_temperature :  string
        Dictionary key containing the toiling temperature of the component (K)

    mol_weight : string
        Dictionary key containing the molecular weight of the component
        (kg/mol)

    temperature : string
        The dictionary key containing the temperature values (K)

    critical_temperature : string
        The dictionary key containing the critical temperature values (K)

    Returns
    -------
    value : NumPy ndarray
        Array containing thermal conductivity values in [W/m.K]

    """
    T = target[temperature]
    Tc = target[critical_temperature]
    MW = target[mol_weight]
    Tbr = target[boiling_temperature] / Tc
    Tr = T / Tc
    value = (
        (1.11 / ((MW * 1e3) ** 0.5))
        * (3 + 20 * (1 - Tr) ** (2 / 3))
        / (3 + 20 * (1 - Tbr) ** (2 / 3))
    )
    return value



r"""

Pore seed models are use to calculate random numbers for each pore, which can
subsequently be used in statistical distributions to calculate actual pore
sizes.

.. autofunction:: openpnm.models.geometry.pore_seed.random
.. autofunction:: openpnm.models.geometry.pore_seed.spatially_correlated

"""
import numpy as _np
import scipy as _sp
from openpnm.models import misc as _misc
from openpnm.utils import logging as _logging
_logger = _logging.getLogger(__name__)


def random(target, seed=None, num_range=[0, 1]):
    return _misc.random(target, element='pore', seed=seed, num_range=num_range)


random.__doc__ = _misc.random.__doc__


def spatially_correlated(target, weights=None, strel=None):
    r"""
    Generates pore seeds that are spatailly correlated with their neighbors.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    weights : list of ints, optional
        The [Nx,Ny,Nz] distances (in number of pores) in each direction that
        should be correlated.

    strel : array_like, optional (in place of weights)
        The option allows full control over the spatial correlation pattern by
        specifying the structuring element to be used in the convolution.

        The array should be a 3D array containing the strength of correlations
        in each direction.  Nonzero values indicate the strength, direction
        and extent of correlations.  The following would achieve a basic
        correlation in the z-direction:

    ::

        strel = np.array([[[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                          [[0, 0, 0], [1, 1, 1], [0, 0, 0]],
                          [[0, 0, 0], [0, 0, 0], [0, 0, 0]]])

    Returns
    -------
    values : NumPy ndarray
        Array containing pore seed values.

    Notes
    -----
    This approach uses image convolution to replace each pore seed in the
    geoemtry with a weighted average of those around it.  It then converts the
    new seeds back to a random distribution by assuming they new seeds are
    normally distributed.

    Because is uses image analysis tools, it only works on Cubic networks.

    This is the appproached used by Gostick et al [2]_ to create an anistropic
    gas diffusion layer for fuel cell electrodes.

    References
    ----------
    .. [2] J. Gostick et al, Pore network modeling of fibrous gas diffusion
           layers for polymer electrolyte membrane fuel cells. J Power Sources
           v173, pp277–290 (2007)

    Examples
    --------
    >>> import openpnm as op
    >>> pn = op.network.Cubic(shape=[10, 10, 10])
    >>> Ps, Ts = pn.Ps, pn.Ts
    >>> geom = op.geometry.GenericGeometry(network=pn, pores=Ps, throats=Ts)
    >>> mod = op.models.geometry.pore_seed.spatially_correlated
    >>> geom.add_model(propname='pore.seed', model=mod, weights=[2, 2, 2])

    """
    import scipy.ndimage as spim
    network = target.project.network
    # The following will only work on Cubic networks
    x = network._shape[0]
    y = network._shape[1]
    z = network._shape[2]
    im = _np.random.rand(x, y, z)
    if strel is None:  # Then generate a strel
        if sum(weights) == 0:
            # If weights of 0 are sent, then skip everything and return rands.
            return im.flatten()
        w = _np.array(weights)
        strel = _np.zeros(w*2+1)
        strel[:, w[1], w[2]] = 1
        strel[w[0], :, w[2]] = 1
        strel[w[0], w[1], :] = 1
    im = spim.convolve(im, strel)
    # Convolution is no longer randomly distributed, so fit a gaussian
    # and find it's seeds
    im = (im - _np.mean(im))/_np.std(im)
    im = 1/2*_sp.special.erfc(-im/_np.sqrt(2))
    values = im.flatten()
    values = values[network.pores(target.name)]
    return values



import inspect
import warnings
import functools
import numpy as _np
import scipy as _sp
import time as _time
import copy
from collections import OrderedDict
from docrep import DocstringProcessor


class Docorator(DocstringProcessor):

    __instance__ = None

    def __new__(cls, *args, **kwargs):
        if Docorator.__instance__ is None:
            Docorator.__instance__ = DocstringProcessor()
        return Docorator.__instance__


class PrintableList(list):
    r"""
    Simple subclass of ``list`` that has nice printing.  Only works flat lists.

    Example
    -------
    >>> from openpnm.utils import PrintableList
    >>> temp = ['item1', 'item2', 'item3']
    >>> print(PrintableList(temp))
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    1     : item1
    2     : item2
    3     : item3
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    Each line contains the result of ``print(item)`` on each item in the list

    """

    def __str__(self):
        horizontal_rule = "―" * 78
        lines = [horizontal_rule]
        self.sort()
        for i, item in enumerate(self):
            lines.append("{0:<5s} : {1}".format(str(i + 1), item))
        lines.append(horizontal_rule)
        return "\n".join(lines)

    def __repr__(self):
        self.sort()
        return super().__repr__()


class PrintableDict(OrderedDict):
    r"""
    Simple subclass of ``dict`` that has nicer printing.

    Example
    -------
    >>> from openpnm.utils import PrintableDict
    >>> from numpy import array as arr
    >>> d = {'item1': 1, 'item2': '1', 'item3': [1, 1], 'item4': arr([1, 1])}
    >>> print(PrintableDict(d))
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    key                                 value
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    item1                               1
    item2                               1
    item3                               [1, 1]
    item4                               (2,)
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    If the item is a Numpy array the value column will contain the items'
    shape, otherwise it will contain the result of ``print(item)``

    """

    def __init__(self, *args, **kwargs):
        self._value = "value"
        self._key = "key"
        super().__init__(*args, **kwargs)

    def __repr__(self):
        text = dict(self).__str__()
        return text

    def __str__(self):
        header = "―" * 78
        lines = [header, "{0:<35s} {1}".format(self._key, self._value), header]
        for item in list(self.keys()):
            if type(self[item]) == _sp.ndarray:
                lines.append("{0:<35s} {1}".format(item, _np.shape(self[item])))
            else:
                lines.append("{0:<35s} {1}".format(item, self[item]))
        lines.append(header)
        return "\n".join(lines)


class SettingsDict(PrintableDict):
    r"""
    The SettingsDict implements the __missing__ magic method, which returns
    None instead of KeyError.  This is useful for checking the value of a
    settings without first ensuring it exists.

    Examples
    --------
    >>> from openpnm.utils import SettingsDict
    >>> sd = SettingsDict()
    >>> sd['test'] = True
    >>> print(sd['test'])
    True
    >>> print(sd['not_a_valid_key'])
    None

    """
    __doc__ = ''

    def __setitem__(self, key, value):
        if hasattr(value, "Np"):
            raise Exception(
                "Cannot store OpenPNM objects in settings, "
                + "store object's name instead"
            )
        super().__setitem__(key, value)

    def __missing__(self, key):
        self[key] = None
        return self[key]

    def _update_settings_and_docs(self, dc):
        if type(dc) is type:  # If dc is class then instantiate it
            dc = dc()
        self.__doc__ = dc.__doc__
        # if dc is a dataclass object.  This step is only necessary to support
        # Python 3.6 which doesn't have the dataclasses module
        if hasattr(dc, '__dict__'):
            dc = copy.deepcopy(dc.__dict__)
        else:
            dc = copy.deepcopy(dc)
        for item in dc.keys():
            self[item] = dc[item]


class GenericSettings:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for item in dir(self):
            if not item.startswith('__'):
                self.__dict__[item] = getattr(self, item)


class SubDict(dict):
    def __getitem__(self, key):
        for item in self.keys():
            if item.endswith('.' + key):
                key = item
        return super().__getitem__(key)


class NestedDict(dict):
    def __init__(self, mapping={}, delimiter="/"):
        super().__init__()
        self.delimiter = delimiter
        self.update(mapping)
        self.unravel()

    def __setitem__(self, key, value):
        path = key.split(self.delimiter, 1)
        if len(path) > 1:
            if path[0] not in self.keys():
                self[path[0]] = NestedDict(delimiter=self.delimiter)
            self[path[0]][path[1]] = value
        else:
            super().__setitem__(key, value)

    def __missing__(self, key):
        self[key] = NestedDict(delimiter=self.delimiter)
        return self[key]

    def unravel(self):
        for item in self.keys():
            self[item] = self.pop(item)

    def to_dict(self, dct=None):
        if dct is None:
            dct = self
        plain_dict = dict()
        for key in dct.keys():
            value = dct[key]
            if hasattr(value, "keys"):
                plain_dict[key] = self.to_dict(value)
            else:
                plain_dict[key] = value
        return plain_dict

    def keys(self, dicts=True, values=True):
        k = list(super().keys())
        new_keys = []
        for item in k:
            if hasattr(self[item], "keys"):
                if dicts:
                    new_keys.append(item)
            else:
                if values:
                    new_keys.append(item)
        return new_keys

    def __str__(self):
        def print_level(self, p="", indent="-"):
            for item in self.keys():
                if hasattr(self[item], "keys"):
                    p = print_level(self[item], p=p, indent=indent + indent[0])
                elif indent[-1] != " ":
                    indent = indent + ""
                p = indent + item + "\n" + p
            return p

        p = print_level(self)
        return p


class HealthDict(PrintableDict):
    r"""
    This class adds a 'health' check to a standard dictionary.  This check
    looks into the dict values, and considers empty lists as healthy and all
    else as unhealthy.  If one or more entries is 'unhealthy' the health method
    returns False.

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_health(self):
        health = True
        for item in list(self.keys()):
            try:
                if len(self[item]) > 0:
                    health = False
            except TypeError:
                if self[item]:
                    health = False
        return health

    health = property(fget=_get_health)


def tic():
    r"""
    Homemade version of matlab tic and toc function, tic starts or resets
    the clock, toc reports the time since the last call of tic.

    See Also
    --------
    toc

    """
    global _startTime_for_tictoc
    _startTime_for_tictoc = _time.time()


def toc(quiet=False):
    r"""
    Homemade version of matlab tic and toc function, tic starts or resets
    the clock, toc reports the time since the last call of tic.

    Parameters
    ----------
    quiet : Boolean
        If False (default) then a message is output to the console.  If True
        the message is not displayed and the elapsed time is returned.

    See Also
    --------
    tic

    """
    if "_startTime_for_tictoc" in globals():
        t = _time.time() - _startTime_for_tictoc
        if quiet is False:
            print(f"Elapsed time in seconds: {t:0.2f}")
        else:
            return t
    else:
        raise Exception("Start time not set, call tic first")


def unique_list(input_list):
    r"""
    For a given list (of points) remove any duplicates

    """
    output_list = []
    if len(input_list) > 0:
        dim = _np.shape(input_list)[1]
        for i in input_list:
            match = False
            for j in output_list:
                if dim == 3:
                    if i[0] == j[0] and i[1] == j[1] and i[2] == j[2]:
                        match = True
                elif dim == 2:
                    if i[0] == j[0] and i[1] == j[1]:
                        match = True
                elif dim == 1:
                    if i[0] == j[0]:
                        match = True
            if match is False:
                output_list.append(i)
    return output_list


def flat_list(input_list):
    r"""
    Given a list of nested lists of arbitrary depth, returns a single level or
    'flat' list.

    """
    x = input_list
    if isinstance(x, list):
        return [a for i in x for a in flat_list(i)]
    else:
        return [x]


def sanitize_dict(input_dict):
    r"""
    Given a nested dictionary, ensures that all nested dicts are normal
    Python dicts.  This is necessary for pickling, or just converting
    an 'auto-vivifying' dict to something that acts normal.

    """
    plain_dict = dict()
    for key in input_dict.keys():
        value = input_dict[key]
        if hasattr(value, "keys"):
            plain_dict[key] = sanitize_dict(value)
        else:
            plain_dict[key] = value
    return plain_dict


def methods_to_table(obj):
    r"""
    """
    parent = obj.__class__.__mro__[1]
    temp = inspect.getmembers(parent, predicate=inspect.isroutine)
    parent_funcs = [i[0] for i in temp if not i[0].startswith("_")]

    temp = inspect.getmembers(obj.__class__, predicate=inspect.isroutine)
    obj_funcs = [i[0] for i in temp if not i[0].startswith("_")]
    funcs = set(obj_funcs).difference(set(parent_funcs))

    row = "+" + "-" * 22 + "+" + "-" * 49 + "+"
    fmt = "{0:1s} {1:20s} {2:1s} {3:47s} {4:1s}"
    lines = []
    lines.append(row)
    lines.append(fmt.format("|", "Method", "|", "Description", "|"))
    lines.append(row.replace("-", "="))
    for i, item in enumerate(funcs):
        try:
            s = getattr(obj, item).__doc__.strip()
            end = s.find("\n")
            if end > 47:
                s = s[:44] + "..."
            lines.append(fmt.format("|", item, "|", s[:end], "|"))
            lines.append(row)
        except AttributeError:
            pass
    return "\n".join(lines)


def models_to_table(obj, params=True):
    r"""
    Converts a ModelsDict object to a ReST compatible table

    Parameters
    ----------
    obj : OpenPNM object
        Any object that has a ``models`` attribute

    params : boolean
        Indicates whether or not to include a list of parameter
        values in the table.  Set to False for just a list of models, and
        True for a more verbose table with all parameter values.

    """
    if not hasattr(obj, "models"):
        raise Exception("Received object does not have any models")
    row = "+" + "-" * 4 + "+" + "-" * 22 + "+" + "-" * 18 + "+" + "-" * 26 + "+"
    fmt = "{0:1s} {1:2s} {2:1s} {3:20s} {4:1s} {5:16s} {6:1s} {7:24s} {8:1s}"
    lines = []
    lines.append(row)
    lines.append(
        fmt.format("|", "#", "|", "Property Name", "|", "Parameter", "|", "Value", "|")
    )
    lines.append(row.replace("-", "="))
    for i, item in enumerate(obj.models.keys()):
        prop = item
        if len(prop) > 20:
            prop = item[:17] + "..."
        temp = obj.models[item].copy()
        model = str(temp.pop("model")).split(" ")[1]
        lines.append(
            fmt.format("|", str(i + 1), "|", prop, "|", "model:", "|", model, "|")
        )
        lines.append(row)
        if params:
            for param in temp.keys():
                p1 = param
                if len(p1) > 16:
                    p1 = p1[:14] + "..."
                p2 = str(temp[param])
                if len(p2) > 24:
                    p2 = p2[:21] + "..."
                lines.append(fmt.format("|", "", "|", "", "|", p1, "|", p2, "|"))
                lines.append(row)
    return "\n".join(lines)


def catch_module_not_found(function):
    """
    A decorator that wraps the passed in function and catches
    ModuleNotFound exception.
    """
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except ModuleNotFoundError:
            pass
    return wrapper


def ignore_warnings(warning=RuntimeWarning):
    r"""
    Decorator for catching warnings. Useful in pore-scale models where nans
    are inevitable, and numpy gets annoying by throwing lots of RuntimeWarnings.

    Parameters
    ----------
    warning : Python Warning object
        Python warning type that you want to temporarily ignore

    Examples
    --------
    >>> from openpnm.utils.misc import ignore_warnings
    >>> @ignore_warnings()
    ... def myfun(x):
    ...     return 1/x

    >>> import numpy as np
    >>> x = np.arange(5)
    >>> myfun(x)
    array([       inf, 1.        , 0.5       , 0.33333333, 0.25      ])

    """

    def _ignore_warning(function):
        @functools.wraps(function)
        def __ignore_warning(*args, **kwargs):
            with warnings.catch_warnings(record=True):
                # Catch all warnings of this type
                warnings.simplefilter("always", warning)
                # Execute the function
                result = function(*args, **kwargs)
            return result

        return __ignore_warning

    return _ignore_warning


def conduit_lengths(network, throats=None, mode="pore"):
    r"""
    Return the respective lengths of the conduit components defined by the throat
    conns P1 - T - P2

    Notes
    -----
    mode = 'pore' - uses pore coordinates
    mode = 'centroid' uses pore and throat centroids

    """
    if throats is None:
        throats = network.throats()
    Ps = network["throat.conns"]
    pdia = network["pore.diameter"]
    Lt = network["throat.length"]

    if mode == "centroid":
        try:
            pcentroids = network["pore.centroid"]
            tcentroids = network["throat.centroid"]
            if _np.sum(_np.isnan(pcentroids)) + _np.sum(_np.isnan(tcentroids)) > 0:
                mode = "pore"
            else:
                plen1 = (
                    _np.sqrt(_np.sum(_sp.square(pcentroids[Ps[:, 0]] - tcentroids), 1))
                    - Lt / 2
                )
                plen2 = (
                    _np.sqrt(_np.sum(_sp.square(pcentroids[Ps[:, 1]] - tcentroids), 1))
                    - Lt / 2
                )
        except KeyError:
            mode = "pore"
    if mode == "pore":
        # Find half-lengths of each pore
        pcoords = network["pore.coords"]
        # Find the pore-to-pore distance, minus the throat length
        lengths = (
            _np.sqrt(_np.sum(_sp.square(pcoords[Ps[:, 0]] - pcoords[Ps[:, 1]]), 1)) - Lt
        )
        lengths[lengths < 0.0] = 2e-9
        # Calculate the fraction of that distance from the first pore
        try:
            fractions = pdia[Ps[:, 0]] / (pdia[Ps[:, 0]] + pdia[Ps[:, 1]])
            # Don't allow zero lengths
            # fractions[fractions == 0.0] = 0.5
            # fractions[fractions == 1.0] = 0.5
        except Exception:
            fractions = 0.5
        plen1 = lengths * fractions
        plen2 = lengths * (1 - fractions)

    return _np.vstack((plen1, Lt, plen2)).T[throats]


def is_symmetric(a, rtol=1e-10):
    r"""
    Is ``a`` a symmetric matrix?

    Parameters
    ----------
    a : ndarray, sparse matrix
        Object to check for being a symmetric matrix.

    rtol : float
        Relative tolerance with respect to the smallest entry in ``a`` that
        is used to determine if ``a`` is symmetric.

    Returns
    -------
    bool
        ``True`` if ``a`` is a symmetric matrix, ``False`` otherwise.

    """
    if type(a) != _sp.ndarray and not _sp.sparse.issparse(a):
        raise Exception("'a' must be either a sparse matrix or an ndarray.")
    if a.shape[0] != a.shape[1]:
        raise Exception("'a' must be a square matrix.")

    atol = _np.amin(_np.absolute(a.data)) * rtol
    if _sp.sparse.issparse(a):
        issym = False if ((a - a.T) > atol).nnz else True
    elif type(a) == _sp.ndarray:
        issym = False if _np.any((a - a.T) > atol) else True

    return issym


def is_valid_propname(propname):
    r"""
    Check if ``propname`` is a valid OpenPNM propname, i.e. starts with
    'pore.' or 'throat.'

    Parameters
    ----------
    propname : str
        Property name to check whether it's a valid OpenPNM propname.

    Returns
    -------
    bool
        Whether or not ``propname`` is a valid name

    """
    if type(propname) is not str:
        return False
    temp = propname.split(".")
    if temp[0] not in ["pore", "throat"]:
        return False
    if len(temp) == 1:
        return False
    for field in temp:
        if len(field) == 0:
            return False
    return True


def nbr_to_str(nbr, t_precision):
    r"""
    Converts a scalar into a string in scientific (exponential) notation
    without the decimal point.

    Parameters
    ----------
    nbr : scalar
        The number to be converted into a scalar.

    t_precision : integer
        The time precision (number of decimal places). Default value is 12.
    """
    from decimal import Decimal as dc
    n = int(-dc(str(round(nbr, t_precision))).as_tuple().exponent
            * (round(nbr, t_precision) != int(nbr)))
    nbr_str = (str(int(round(nbr, t_precision)*10**n)) + ('e-'+str(n))*(n != 0))
    return nbr_str



import inspect
from openpnm.utils import PrintableDict, logging, Workspace
from openpnm.utils.misc import is_valid_propname
logger = logging.getLogger(__name__)
ws = Workspace()


class ModelsDict(PrintableDict):
    r"""
    This subclassed dictionary is assigned to the ``models`` attribute of
    all objects that inherit from the ``ModelsMixin`` class.  Each dictionary
    entry corresponds to an entry in the target object's dictionary, and
    contains the models and associated parameters for generating the model.

    The main features of this subclass are three methods the help resolve the
    order in which models should be called: ``dependency_list``,
    ``dependency_graph``, and ``dependency_map``.

    """
    def dependency_list(self):
        r"""
        Returns a list of dependencies in the order with which they should be
        called to ensure data is calculated by one model before it's asked for
        by another.

        Notes
        -----
        This raises an exception if the graph has cycles which means the
        dependencies are unresolvable (i.e. there is no order which the
        models can be called that will work).  In this case it is possible
        to visually inspect the graph using ``dependency_graph``.

        See Also
        --------
        dependency_graph
        dependency_map

        """
        import networkx as nx

        dtree = self.dependency_graph()
        cycles = list(nx.simple_cycles(dtree))
        if cycles:
            raise Exception('Cyclic dependency found: ' + ' -> '.join(
                            cycles[0] + [cycles[0][0]]))
        d = nx.algorithms.dag.lexicographical_topological_sort(dtree, sorted)
        return list(d)

    def dependency_graph(self, deep=False):
        r"""
        Returns a NetworkX graph object of the dependencies

        Parameters
        ----------
        deep : bool, optional
            Defines whether intra- or inter-object dependency graph is desired.
            Default is False, i.e. only returns dependencies within the object.

        See Also
        --------
        dependency_list
        dependency_map

        Notes
        -----
        To visualize the dependencies, the following NetworkX function and
        settings is helpful:

        nx.draw_spectral(
            dtree,
            arrowsize=50,
            font_size=32,
            with_labels=True,
            node_size=2000,
            width=3.0,
            edge_color='lightgrey',
            font_weight='bold'
        )

        """
        import networkx as nx

        dtree = nx.DiGraph()
        models = list(self.keys())

        for model in models:
            dtree.add_node(model)
            # Filter pore/throat props only
            dependencies = set()
            for param in self[model].values():
                if is_valid_propname(param):
                    dependencies.add(param)
            # Add depenency from model's parameters
            for d in dependencies:
                if not deep:
                    if d in models:
                        dtree.add_edge(d, model)
                else:
                    dtree.add_edge(d, model)

        return dtree

    def dependency_map(self, ax=None, figsize=None):
        r"""
        Create a graph of the dependency graph in a decent format

        Parameters
        ----------
        ax : matplotlib.axis, optional
            Matplotlib axis object on which dependency map is to be drawn

        figsize : tuple, optional
            Tuple containing frame size

        See Also
        --------
        dependency_graph
        dependency_list

        """
        import networkx as nx
        import matplotlib.pyplot as plt

        dtree = self.dependency_graph()
        labels = {}
        for node in dtree.nodes:
            if node.startswith("pore."):
                value = node.replace("pore.", "[P] ")
            elif node.startswith("throat."):
                value = node.replace("throat.", "[T] ")
            labels[node] = value

        if ax is None:
            fig, ax = plt.subplots()
        fig.set_size_inches(figsize)

        nx.draw_shell(
            dtree,
            labels=labels,
            with_labels=True,
            edge_color='lightgrey',
            font_size=12,
            width=3.0,
        )

        ax = plt.gca()
        ax.margins(x=0.2, y=0.02)

        # return ax.figure

    def __str__(self):
        horizontal_rule = '―' * 85
        lines = [horizontal_rule]
        strg = '{0:<3s} {1:<35s} {2:<25s} {3}'
        lines.append(strg.format('#', 'Property Name', 'Parameter', 'Value'))
        lines.append(horizontal_rule)
        for i, item in enumerate(self.keys()):
            temp = self[item].copy()
            regen_mode = temp.pop('regen_mode', None)
            model = str(temp.pop('model')).split(' ')[1]
            lines.append(strg.format(str(i+1), item, 'model:', model))
            for param in temp.keys():
                lines.append(strg.format('', '', param+':', temp[param]))
            lines.append(strg.format('', '', 'regeneration mode:', regen_mode))
            lines.append(horizontal_rule)
        return '\n'.join(lines)


class ModelWrapper(dict):
    r"""
    This class is used to hold individual models and provide some extra
    functionality, such as pretty-printing.
    """
    @property
    def propname(self):
        for proj in ws.values():
            for obj in proj:
                if hasattr(obj, 'models'):
                    for key, mod in obj.models.items():
                        if mod is self:
                            return key

    def __str__(self):
        horizontal_rule = '―' * 78
        lines = [horizontal_rule]
        strg = '{0:<25s} {2:<25s} {2}'
        lines.append(strg.format('Property Name', 'Parameter', 'Value'))
        lines.append(horizontal_rule)
        temp = self.copy()
        regen_mode = temp.pop('regen_mode', None)
        model = str(temp.pop('model')).split(' ')[1]
        lines.append(strg.format(self.propname, 'model:', model))
        for param in temp.keys():
            lines.append(strg.format('', param+':', temp[param]))
        lines.append(strg.format('', 'regeneration mode:', regen_mode))
        lines.append(horizontal_rule)
        return '\n'.join(lines)


class ModelsMixin:
    r"""
    This class is meant to be combined by the Base class in multiple
    inheritence.  This approach is used since Network and Algorithm do not
    need to have any ``models`` attribute, while Phase, Geometry, and Physics
    do.  By using a mixin class, all objects can inherit from Base while
    the model functionality can be added only where needed.

    Notes
    -----
    The following table gives a brief overview of the methods that are added
    to the object by this mixin.  In addition to these methods, a ``models``
    attribute is also added, which is a dictionary that contains all of the
    models and their parameters.

    +----------------------+--------------------------------------------------+
    | Method or Attribute  | Functionality                                    |
    +======================+==================================================+
    | ``add_model``        | Add a given model and parameters to the object   |
    +----------------------+--------------------------------------------------+
    | ``regenerate_model`` | Runs the model(s) to recalculate data            |
    +----------------------+--------------------------------------------------+
    | ``remove_model``     | Removes specified model as well as it's data     |
    +----------------------+--------------------------------------------------+

    Examples
    --------
    >>> import openpnm as op

    Create a Demo class using Base and ModelsMixin:

    >>> class Demo(op.core.Base, op.core.ModelsMixin):
    ...     pass
    >>> temp = Demo(Np=3, Nt=2)

    The new class has the normal Base methods:

    >>> print(temp.num_pores())
    3

    But also has those needed for working with models.  For instance, a simple
    model can be added as follows:

    >>> temp.add_model(propname='pore.test',
    ...                model=op.models.misc.constant,
    ...                value=2)
    >>> print(temp['pore.test'])
    [2 2 2]

    All the models and their respective parameters are stored in the ``models``
    attribute:

    >>> print(temp.models)
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    #   Property Name                       Parameter                 Value
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    1   pore.test                           model:                    constant
                                            value:                    2
                                            regeneration mode:        normal
    ―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    """

    def add_model(self, propname, model, regen_mode='', **kwargs):
        r"""
        Adds a new model to the models dictionary (``object.models``)

        Parameters
        ----------
        propname : string
            The name of the property to be calculated by the model.

        model : function
            A reference (handle) to the function to be used.

        regen_mode : string
            Controls how/when the model is run (See Notes for more details).
            Options are:

            *'normal'* : (default) The model is run directly upon being
            assiged, and also run every time ``regenerate_models`` is called.

            *'constant'* : The model is run directly upon being assigned, but
            is not called again, thus making it's data act like a constant.
            If, however, the data is deleted from the object it will be
            regenerated again.

            *'deferred'* Is not run upon being assigned, but is run the first
            time that ``regenerate_models`` is called.

        """
        if propname in kwargs.values():  # Prevent infinite loops of look-ups
            raise Exception(propname+' can\'t be both dependency and propname')
        # Look for default regen_mode in settings if present, else use 'normal'
        if regen_mode == '':
            if 'regen_mode' in self.settings.keys():
                regen_mode = self.settings['regen_mode']
            else:
                regen_mode = 'normal'
        # Add model and regen_mode to kwargs dictionary
        kwargs.update({'model': model, 'regen_mode': regen_mode})
        # Insepct model to extract arguments and default values
        if model.__defaults__:
            vals = list(inspect.getfullargspec(model).defaults)
            keys = inspect.getfullargspec(model).args[-len(vals):]
            for k, v in zip(keys, vals):  # Put defaults into kwargs
                if k not in kwargs:  # Skip if argument was given in kwargs
                    kwargs.update({k: v})
        self.models[propname] = ModelWrapper(kwargs)  # Store all kwargs
        # Regenerate model values if necessary
        if regen_mode not in ['deferred', 'explicit']:
            self._regen(propname)

    def regenerate_models(self, propnames=None, exclude=[], deep=False):
        r"""
        Re-runs the specified model or models.

        Parameters
        ----------
        propnames : string or list of strings
            The list of property names to be regenerated.  If None are given
            then ALL models are re-run (except for those whose ``regen_mode``
            is 'constant').

        exclude : list of strings
            Since the default behavior is to run ALL models, this can be used
            to exclude specific models.  It may be more convenient to supply
            as list of 2 models to exclude than to specify 8 models to include.

        deep : boolean
            Specifies whether or not to regenerate models on all associated
            objects.  For instance, if ``True``, then all Physics models will
            be regenerated when method is called on the corresponding Phase.
            The default is ``False``.  The method does not work in reverse,
            so regenerating models on a Physics will not update a Phase.

        """
        # If empty list of propnames was given, do nothing and return
        if type(propnames) is list and len(propnames) == 0:
            return
        if type(propnames) is str:  # Convert string to list if necessary
            propnames = [propnames]
        if propnames is None:  # If no props given, then regenerate them all
            propnames = self.models.dependency_list()
            # If some props are to be excluded, remove them from list
            exclude.extend([k for k, v in self.models.items()
                            if v['regen_mode'] == 'explicit'])
            propnames = [i for i in propnames if i not in exclude]
        # Re-order given propnames according to dependency tree
        self_models = self.models.dependency_list()
        propnames = [i for i in self_models if i in propnames]

        if deep:
            other_models = None  # Will trigger regen of ALL models
        else:
            # Make list of given propnames that are not in self
            other_models = list(set(propnames).difference(set(self_models)))
        # The following has some redundant lines, but is easier to understand
        if self._isa('phase'):
            # Start be regenerating models on self
            for item in propnames:
                self._regen(item)
            # Then regen models on associated objects, if any in other_models
            for phys in self.project.find_physics(phase=self):
                phys.regenerate_models(propnames=other_models, deep=False)
        elif self._isa('network'):  # Repeat for other object types
            for item in propnames:
                self._regen(item)
            for geom in self.project.geometries().values():
                geom.regenerate_models(propnames=other_models, deep=False)
        else:
            for item in propnames:
                self._regen(item)

    def _regen(self, prop):
        # Create a temporary dict of all model arguments
        try:
            kwargs = self.models[prop].copy()
        except KeyError:
            logger.info(prop+' not found, will retry if deep is True')
            return
        # Pop model and regen_mode from temporary dict
        model = kwargs.pop('model')
        regen_mode = kwargs.pop('regen_mode', None)
        # Only regenerate model if regen_mode is correct
        if self.settings['freeze_models']:
            # Don't run ANY models if freeze_models is set to True
            pass
        elif regen_mode == 'constant':
            # Only regenerate if data not already in dictionary
            if prop not in self.keys():
                self[prop] = model(target=self, **kwargs)
        else:
            try:
                self[prop] = model(target=self, **kwargs)
            except KeyError as e:
                logger.error(prop + ' was not run since the following '
                             + 'property is missing: ' + e.__str__())
                self.models[prop]['regen_mode'] = 'deferred'

    def remove_model(self, propname=None, mode=['model', 'data']):
        r"""
        Removes model and data from object.

        Parameters
        ----------
        propname : string or list of strings
            The property or list of properties to remove

        mode : list of strings
            Controls what is removed.  Options are:

            *'model'* : Removes the model but not any numerical data that may
            already exist.

            *'data'* : Removes the data but leaves the model.

        The default is both.

        """
        if type(propname) is str:
            propname = [propname]
        for item in propname:
            if 'model' in mode:
                if item in self.models.keys():
                    del self.models[item]
            if 'data' in mode:
                if item in self.keys():
                    del self[item]

    def _get_models(self):
        if not hasattr(self, '_models_dict'):
            self._models_dict = ModelsDict()
        return self._models_dict

    def _set_models(self, dict_):
        self._models_dict = ModelsDict()
        # Renerate all models in new dict if regen mode says so
        for model in dict_.keys():
            self.add_model(propname=model, **dict_[model])

    models = property(fget=_get_models, fset=_set_models)



import numpy as np
import scipy.sparse as sprs
from decimal import Decimal as dc
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging, GenericSettings, Docorator
docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.get_sectionsf('TransientReactiveTransportSettings',
                      sections=['Parameters', 'Other Parameters'])
@docstr.dedent
class TransientReactiveTransportSettings(GenericSettings):
    r"""

    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s

    quantity : (str)
        The name of the physical quantity to be calculated
    conductance : (str)
        The name of the pore-scale transport conductance values. These are
        typically calculated by a model attached to a *Physics* object
        associated with the given *Phase*.

    Other Parameters
    ----------------
    t_initial : scalar
        The simulation's start time, which must be smaller than
        't_final'. The default value is 0.
    t_final : scalar
        The simulation's end time, which must be bigger than 't_initial'.
        The default value is 10.
    t_step : scalar
        The simulation's time step, which must be smaller than 't_final' less
        and 't_initial'. The default value is 0.1.
    t_output : scalar, ND-array, or list
        When 't_output' is a scalar, it is considered as an output interval
        to store transient solutions. The default value is 1e+08. Initial,
        final and steady-state (if reached) fields are always stored. If
        't_output' > 't_final', no transient data is stored. If 't_output'
        is not a multiple of 't_step', 't_output' will be approximated.
        When 't_output' is a list or ND-array, transient solutions
        corresponding to this list or array will be stored.
    output_times : list
        List of output times. The values in the list must be multiples of
        the time step 't_step'.
    t_tolerance : scalar
        Transient solver tolerance. The simulation stops (before reaching
        't_final') when the residual falls below 't_tolerance'. The
        default value is 1e-06. The 'residual' measures the variation from
        one time-step to another in the value of the 'quantity' solved for.
    t_precision : integer
        The time precision (number of decimal places).
    t_scheme : string
        The time discretization scheme. Three options available: 'steady'
        to perform a steady-state simulation, and 'implicit' (fast, 1st
        order accurate) and 'cranknicolson' (slow, 2nd order accurate) both
        for transient simulations. The default value is 'implicit'.

    ----

    **The following parameters pertain to the ReactiveTransport class**

    %(ReactiveTransportSettings.other_parameters)s

    ----

    **The following parameters pertain to the GenericTransport class**

    %(GenericTransportSettings.other_parameters)s

    """

    phase = None
    t_initial = 0
    t_final = 10
    t_step = 0.1
    t_output = 1e+08
    t_tolerance = 1e-06
    t_precision = 12
    t_scheme = 'implicit'


class TransientReactiveTransport(ReactiveTransport):
    r"""
    A subclass of ReactiveTransport for transient/steady-state simulations

    Parameters
    ----------
    network : OpenPNM Network object
        The Network with which this algorithm is associated.

    project : OpenPNM Project object
        Either a Network or a Project must be specified.

    Notes
    -----

    This subclass performs steady and transient simulations of transport
    phenomena with reactions when source terms are added. It supports 3 time
    discretization schemes; 'steady' to perform a steady-state simulation, and
    'implicit' (fast, 1st order accurate) and 'cranknicolson' (slow, 2nd order
    accurate) both for transient simulations.
    """

    def __init__(self, settings={}, phase=None, **kwargs):
        super().__init__(**kwargs)
        self.settings._update_settings_and_docs(TransientReactiveTransportSettings)
        self.settings.update(settings)
        # Initialize the steady sys of eqs A matrix
        self._A_steady = None
        if phase is not None:
            self.setup(phase=phase)
        # Initialize the initial condition
        self["pore.ic"] = np.nan

    def setup(self, phase=None, quantity='', conductance='',
              t_initial=None, t_final=None, t_step=None, t_output=None,
              t_tolerance=None, t_precision=None, t_scheme='', **kwargs):
        r"""
        This method takes several arguments that are essential to running the
        algorithm and adds them to the settings

        Parameters
        ----------

        Notes
        -----
        More settings can be adjusted in the presence of a non-linear source
        term such as under-relaxation.
        See the 'ReactiveTransport' class documentation for details.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        if t_initial is not None:
            self.settings['t_initial'] = t_initial
        if t_final is not None:
            self.settings['t_final'] = t_final
        if t_step is not None:
            self.settings['t_step'] = t_step
        if t_output is not None:
            self.settings['t_output'] = t_output
        if t_tolerance is not None:
            self.settings['t_tolerance'] = t_tolerance
        if t_precision is not None:
            self.settings['t_precision'] = t_precision
        if t_scheme:
            self.settings['t_scheme'] = t_scheme
        self.settings.update(kwargs)

    def set_IC(self, values):
        r"""
        A method to set simulation initial conditions

        Parameters
        ----------
        values : ND-array or scalar
            Set the initial conditions using an 'Np' long array. 'Np' being
            the number of pores. If a scalar is given, the same value is
            imposed to all pores.

        """
        quantity = self.settings['quantity']
        values = np.array(values) * 1.0
        if values.size > 1 and values.size != self.Np:
            raise Exception('The number of initial values must be either 1 or Np')
        self['pore.ic'] = values
        if not quantity:
            raise Exception('"quantity" has not been defined on this algorithm')
        self[quantity] = values

    def _get_f1_f2_f3(self):
        r"""
        Helper method: returns f1, f2, and f3 for _t_update_A and _t_update_b methods.
        """
        s = self.settings['t_scheme']
        if s == 'implicit':
            f1, f2, f3 = 1, 1, 0
        elif s == 'cranknicolson':
            f1, f2, f3 = 0.5, 1, 0
        elif s == 'steady':
            f1, f2, f3 = 1, 0, 1
        else:
            raise Exception(f'Unsupported t_scheme: "{s}"')
        return f1, f2, f3

    def _t_update_A(self):
        r"""
        A method to update 'A' matrix at each time step according to 't_scheme'
        """
        network = self.project.network
        Vi = network['pore.volume']
        dt = self.settings['t_step']
        f1, f2, _ = self._get_f1_f2_f3()
        # Compute A (operations involve conversion to 'csr')
        A = ((f2/dt) * sprs.coo_matrix.multiply(
            sprs.coo_matrix(np.reshape(Vi, (self.Np, 1)), shape=(self.Np, 1)),
            sprs.identity(self.Np, format='coo')) + f1 * self._A_steady)
        # Convert A to 'coo' format to apply BCs
        A = sprs.coo_matrix(A)
        self._A = A
        return A

    def _t_update_b(self):
        r"""
        A method to update 'b' array at each time step according to
        't_scheme' and the source term value
        """
        quantity = self.settings['quantity']
        network = self.project.network
        phase = self.project.phases()[self.settings['phase']]
        Vi = network['pore.volume']
        dt = self.settings['t_step']
        f1, f2, f3 = self._get_f1_f2_f3()
        x_old = self[quantity]
        b = (f2 * (1-f1) * (-self._A_steady) * x_old
             + f2 * (Vi/dt) * x_old
             + f3 * np.zeros(shape=(self.Np,), dtype=float))
        self._update_iterative_props()
        for item in self.settings['sources']:
            Ps = self.pores(item)
            # Update b
            b[Ps] = b[Ps] - f2 * (1-f1) * (phase[item + '.' + 'rate'][Ps])
        self._b = b
        return b

    def run(self, t=None):
        r"""
        Builds 'A' matrix of the steady system of equations to be used at each
        time step to build transient 'A' and 'b'. Imposes the initial
        conditions and stores the initial field. Initialize transient 'A', 'b',
        and source term (if present) and finally calls the transient solver.

        Parameters
        ----------
        t : scalar
            The time to start the simulation from. If no time is specified, the
            simulation starts from 't_initial' defined in the settings.
        """
        logger.info('―' * 80)
        logger.info('Running TransientTransport')
        # If ICs are not defined, assume zero
        if not np.isfinite(self["pore.ic"]).all():
            self.set_IC(0)
        # Make sure _A is None to force _build_A, otherwise _A_steady might be wrong
        self._A = None
        # Save A matrix of the steady state problem (without BCs applied)
        self._A_steady = self.A.copy()
        # Initialize A and b with BCs applied
        self._t_update_A()
        self._t_update_b()
        self._apply_BCs()
        # Save copies of A and b to be used in _t_run_reactive()
        self._A_t = self._A.copy()
        self._b_t = self._b.copy()
        if t is None:
            t = self.settings['t_initial']
        # Create S1 & S2 for 1st Picard's iteration
        self._update_iterative_props()
        self._run_transient(t=t)

    def _run_transient(self, t):
        """r
        Performs a transient simulation according to the specified settings
        updating 'b' and calling '_t_run_reactive' at each time step.
        Stops after reaching the end time 't_final' or after achieving the
        specified tolerance 't_tolerance'. Stores the initial and steady-state
        (if obtained) fields in addition to transient data (according to the
        specified 't_output').

        Parameters
        ----------
        t : scalar
            The time to start the simulation from.

        Notes
        -----
        Transient solutions are stored on the object under
        ``pore.quantity_timeStepIndex`` where *quantity* is specified in the
        ``settings`` attribute. Initial field is stored as
        ``pore.quantity_initial``. Steady-state solution (if reached) is stored
        as ``pore.quantity_steady``. Current solution is stored as
        ``pore.quantity``.
        """
        tf = self.settings['t_final']
        dt = self.settings['t_step']
        to = self.settings['t_output']
        tol = self.settings['t_tolerance']
        t_pre = self.settings['t_precision']
        quantity = self.settings['quantity']
        s = self.settings['t_scheme']
        res_t = 1e+06  # Initialize the residual

        if type(to) in [float, int]:
            # Make sure 'tf' and 'to' are multiples of 'dt'
            tf = tf + (dt-(tf % dt))*((tf % dt) != 0)
            to = to + (dt-(to % dt))*((to % dt) != 0)
            self.settings['t_final'] = tf
            self.settings['t_output'] = to
            out = np.arange(t+to, tf, to)
        elif type(to) in [np.ndarray, list]:
            out = np.array(to)
        out = np.append(out, tf)
        out = np.unique(out)
        out = np.around(out, decimals=t_pre)

        # If solver in steady mode, do one iteration
        if s == 'steady':
            logger.info('    Running in steady mode')
            self._t_run_reactive()

        # Time marching step
        else:
            # Export the initial field (t=t_initial)
            t_str = self._nbr_to_str(t)
            quant_init = self["pore.ic"]
            self[quantity + '@' + t_str] = quant_init
            self[quantity] = quant_init

            for time in np.arange(t+dt, tf+dt, dt):
                if res_t >= tol:  # Check if the steady state is reached
                    logger.info(f'    Current time step: {time} s')
                    # Update A and b and apply BCs
                    self._t_update_A()
                    self._t_update_b()
                    self._apply_BCs()
                    # Save copies of A and b to be used in _t_run_reactive()
                    self._A_t = self._A.copy()
                    self._b_t = self._b.copy()
                    x_old = self[quantity]
                    self._t_run_reactive(x0=x_old)
                    x_new = self[quantity]
                    # Compute the residual
                    res_t = np.sum(np.absolute(x_old**2 - x_new**2))
                    logger.info(f'        Residual: {res_t}')
                    # Output transient solutions. Round time to ensure every
                    # value in outputs is exported.
                    if round(time, t_pre) in out:
                        t_str = self._nbr_to_str(time)
                        self[quantity + '@' + t_str] = x_new
                        logger.info(f'        Exporting time step: {time} s')
                else:
                    # Output steady state solution
                    t_str = self._nbr_to_str(time)
                    self[quantity + '@' + t_str] = x_new
                    logger.info(f'        Exporting time step: {time} s')
                    break

            if round(time, t_pre) == tf:
                logger.info(f'    Maximum time step reached: {time} s')
            else:
                logger.info(f'    Transient solver converged after: {time} s')

    def _t_run_reactive(self, x0=None):
        """r
        Repeatedly updates transient 'A', 'b', and the solution guess within
        each time step according to the applied source term then calls '_solve'
        to solve the resulting system of linear equations. Stops when the
        residual falls below 'solver_tol'.

        Parameters
        ----------
        x0 : ND-array
            Initial guess of unknown variable

        Returns
        -------
        x_new : ND-array
            Solution array.

        Notes
        -----
        Description of 'relaxation_quantity' and 'nlin_max_iter' settings can be
        found in the parent class 'ReactiveTransport' documentation.

        """
        quantity = self.settings['quantity']
        w = self.settings['relaxation_quantity']
        max_it = int(self.settings['nlin_max_iter'])
        x = np.zeros(self.Np, dtype=float) if x0 is None else x0.copy()

        # Write initial guess to algorithm for _update_iterative_props to work
        self[quantity] = x

        for itr in range(max_it):
            # Update iterative properties on phase and physics
            self._update_iterative_props()
            # Build A and b, apply source terms and correct according to scheme
            self._A = self._A_t.copy()
            self._b = self._b_t.copy()
            self._apply_sources()
            self._correct_apply_sources()
            # Compute the residual
            res = self._get_residual()
            if itr >= 1 and self._is_converged():
                logger.info(f'Solution converged: {res:.4e}')
                return x
            logger.info(f'Tolerance not met: {res:.4e}')
            # Solve, use relaxation, and update solution on algorithm obj
            self[quantity] = x = self._solve(x0=x) * w + x * (1 - w)

        # Check solution convergence after max_it iterations
        if not self._is_converged():
            raise Exception(f"Not converged after {max_it} iterations.")

    def results(self, times=None, **kwargs):
        r"""
        Fetches the calculated quantity from the algorithm and returns it as
        an array.

        Parameters
        ----------
        times : scalar, ND-array, list of scalars, None, or string
            Time steps to be returned. The default value is None which results
            in returning all time steps. If times is a scalar, only the
            corresponding time step is returned. If times is an ND-array or a
            list of scalars, time steps in the provided array or list are
            returned. If times is 'final' or 'actual', the current value of the
            quantity is returned.

        t_precision : integer
            The time precision (number of decimal places). Default value is 12.

        Notes
        -----
        The keyword steps is interpreted in the same way as times.
        """
        if 'steps' in kwargs.keys():
            times = kwargs['steps']
        t_pre = self.settings['t_precision']
        quantity = self.settings['quantity']
        q = [k for k in list(self.keys()) if quantity in k]
        if times is None:
            t = q
        elif times in ['final', 'actual']:
            t = [quantity]
        elif type(times) in [np.ndarray, list, float, int]:
            out = np.array(times)
            out = np.unique(out)
            out = np.around(out, decimals=t_pre)
            t = []
            for i in out:
                j = self._nbr_to_str(i)
                t_str = [k for k in q if j == k.split('@')[-1]]
                t += (t_str)
            # Times stored by the transient algorithm
            strd_t = [float(st.split('@')[1]) for st in q if '@' in st]
            strd_t = np.array(strd_t)
            # Times requested but non stored by the algorithm
            missing_t = np.setdiff1d(np.around(out, decimals=t_pre),
                                     np.around(strd_t, decimals=t_pre))
            if missing_t.size != 0:
                logger.warning('Time(s) '+str(missing_t)+' not stored.')
        d = {k: self[k] for k in t}
        return d

    def _nbr_to_str(self, nbr, t_pre=None):
        r"""
        Converts a scalar into a string in scientific (exponential) notation
        without the decimal point.

        Parameters
        ----------
        nbr : scalar
            The number to be converted into a scalar.

        t_precision : integer
            The time precision (number of decimal places). Default value is 12.
        """
        if t_pre is None:
            t_pre = self.settings['t_precision']
        n = int(-dc(str(round(nbr, t_pre))).as_tuple().exponent
                * (round(nbr, t_pre) != int(nbr)))
        nbr_str = (str(int(round(nbr, t_pre)*10**n)) + ('e-'+str(n))*(n != 0))
        return nbr_str

    def _correct_apply_sources(self):
        """r
        Update 'A' and 'b' correcting the already applied source terms to
        specified pores

        Notes
        -----
        Correction (built for transient simulations) depends on the time scheme
        """
        if self.settings['t_scheme'] == 'cranknicolson':
            f1 = 0.5
        else:
            f1 = 1.0
        phase = self.project.phases()[self.settings['phase']]
        for item in self.settings['sources']:
            Ps = self.pores(item)
            # get already added relaxed source term
            S1, S2 = [phase[item + '.' + x][Ps] for x in ['S1', 'S2']]
            # correct S1 and S2 in A and b as a function of t_scheme
            datadiag = self._A.diagonal().copy()
            datadiag[Ps] = datadiag[Ps] - S1 + f1*S1
            self._A.setdiag(datadiag)
            self._b[Ps] = self._b[Ps] + S2 - f1*S2



import numpy as np
import scipy as sp
from openpnm.topotools import trim
from openpnm.utils import logging
from openpnm.io import GenericIO
from openpnm.network import GenericNetwork
from pathlib import Path
logger = logging.getLogger(__name__)


class Statoil(GenericIO):
    r"""
    The StatOil format is used by the Maximal Ball network extraction code of
    the Imperial College London group

    This class can be used to load and work with those networks.  Numerous
    datasets are available for download from the group's
    `website <http://tinyurl.com/zurko4q>`_.

    The so-called 'Statoil' format consists of 4 different files in a single
    folder.  The data is stored in columns with each corresponding to a
    specific property.  Headers are not provided in the files, so one must
    refer to various theses and documents to interpret their meaning.
    """
    @classmethod
    def load(cls, path, prefix, network=None):
        r"""
        Load data from the \'dat\' files located in specified folder.

        Parameters
        ----------
        path : string
            The full path to the folder containing the set of \'dat\' files.

        prefix : string
            The file name prefix on each file. The data files are stored
            as \<prefix\>_node1.dat.

        network : OpenPNM Network Object
            If given then the data will be loaded on it and returned.  If not
            given, a Network will be created and returned.

        Returns
        -------
        An OpenPNM Project containing a GenericNetwork holding all the data

        """
        from pandas import read_table, DataFrame

        net = {}

        # Parse the link1 file
        path = Path(path)
        filename = Path(path.resolve(), prefix+'_link1.dat')
        with open(filename, mode='r') as f:
            link1 = read_table(filepath_or_buffer=f,
                                  header=None,
                                  skiprows=1,
                                  sep=' ',
                                  skipinitialspace=True,
                                  index_col=0)
        link1.columns = ['throat.pore1', 'throat.pore2', 'throat.radius',
                         'throat.shape_factor', 'throat.total_length']
        # Add link1 props to net
        net['throat.conns'] = np.vstack((link1['throat.pore1']-1,
                                         link1['throat.pore2']-1)).T
        net['throat.conns'] = np.sort(net['throat.conns'], axis=1)
        net['throat.radius'] = np.array(link1['throat.radius'])
        net['throat.shape_factor'] = np.array(link1['throat.shape_factor'])
        net['throat.total_length'] = np.array(link1['throat.total_length'])

        filename = Path(path.resolve(), prefix+'_link2.dat')
        with open(filename, mode='r') as f:
            link2 = read_table(filepath_or_buffer=f,
                                  header=None,
                                  sep=' ',
                                  skipinitialspace=True,
                                  index_col=0)
        link2.columns = ['throat.pore1', 'throat.pore2',
                         'throat.pore1_length', 'throat.pore2_length',
                         'throat.length', 'throat.volume',
                         'throat.clay_volume']
        # Add link2 props to net
        cl_t = np.array(link2['throat.length'])
        net['throat.length'] = cl_t
        net['throat.conduit_lengths.throat'] = cl_t
        net['throat.volume'] = np.array(link2['throat.volume'])
        cl_p1 = np.array(link2['throat.pore1_length'])
        net['throat.conduit_lengths.pore1'] = cl_p1
        cl_p2 = np.array(link2['throat.pore2_length'])
        net['throat.conduit_lengths.pore2'] = cl_p2
        net['throat.clay_volume'] = np.array(link2['throat.clay_volume'])
        # ---------------------------------------------------------------------
        # Parse the node1 file
        filename = Path(path.resolve(), prefix+'_node1.dat')
        with open(filename, mode='r') as f:
            row_0 = f.readline().split()
            num_lines = int(row_0[0])
            array = sp.ndarray([num_lines, 6])
            for i in range(num_lines):
                row = f.readline()\
                       .replace('\t', ' ').replace('\n', ' ').split()
                array[i, :] = row[0:6]
        node1 = DataFrame(array[:, [1, 2, 3, 4]])
        node1.columns = ['pore.x_coord', 'pore.y_coord', 'pore.z_coord',
                         'pore.coordination_number']
        # Add node1 props to net
        net['pore.coords'] = np.vstack((node1['pore.x_coord'],
                                        node1['pore.y_coord'],
                                        node1['pore.z_coord'])).T
        # ---------------------------------------------------------------------
        # Parse the node1 file
        filename = Path(path.resolve(), prefix+'_node2.dat')
        with open(filename, mode='r') as f:
            node2 = read_table(filepath_or_buffer=f,
                                  header=None,
                                  sep=' ',
                                  skipinitialspace=True,
                                  index_col=0)
        node2.columns = ['pore.volume', 'pore.radius', 'pore.shape_factor',
                         'pore.clay_volume']
        # Add node2 props to net
        net['pore.volume'] = np.array(node2['pore.volume'])
        net['pore.radius'] = np.array(node2['pore.radius'])
        net['pore.shape_factor'] = np.array(node2['pore.shape_factor'])
        net['pore.clay_volume'] = np.array(node2['pore.clay_volume'])
        net['throat.area'] = ((net['throat.radius']**2)
                              / (4.0*net['throat.shape_factor']))
        net['pore.area'] = ((net['pore.radius']**2)
                            / (4.0*net['pore.shape_factor']))

        if network is None:
            network = GenericNetwork()
        network = cls._update_network(network=network, net=net)

        # Use OpenPNM Tools to clean up network
        # Trim throats connected to 'inlet' or 'outlet' reservoirs
        trim1 = np.where(np.any(net['throat.conns'] == -1, axis=1))[0]
        # Apply 'outlet' label to these pores
        outlets = network['throat.conns'][trim1, 1]
        network['pore.outlets'] = False
        network['pore.outlets'][outlets] = True
        trim2 = np.where(np.any(net['throat.conns'] == -2, axis=1))[0]
        # Apply 'inlet' label to these pores
        inlets = network['throat.conns'][trim2, 1]
        network['pore.inlets'] = False
        network['pore.inlets'][inlets] = True
        # Now trim the throats
        to_trim = np.hstack([trim1, trim2])
        trim(network=network, throats=to_trim)

        return network.project



import pickle
from flatdict import FlatDict
from openpnm.utils import NestedDict, sanitize_dict, Workspace
from openpnm.utils import logging
from openpnm.io import GenericIO
logger = logging.getLogger(__name__)
ws = Workspace()


class Dict(GenericIO):
    r"""
    Generates hierarchical ``dicts`` with a high degree of control over the
    structure.

    This is the most important class in the ``io`` module, since many other
    classes use this to manipulate and format the data structures.

    Also, it is possible to use Python's ``pickle`` module to save ``dicts``
    to file.

    """

    @classmethod
    def from_dict(cls, dct, project=None, delim=' | '):
        r"""
        This method converts a correctly formatted dictionary into OpenPNM
        objects, and returns a handle to the *project* containing them.

        Parameters
        ----------
        dct : dictionary
            The Python dictionary containing the data.  The nesting and
            labeling of the dictionary is used to create the appropriate
            OpenPNM objects.

        project : OpenPNM Project Object
            The project with which the created objects should be associated.
            If not supplied, one will be created.

        Returns
        -------
        An OpenPNM Project containing the objects created to store the given
        data.

        Notes
        -----
        The requirement of a *correctly formed* dictionary is rather strict,
        and essentially means a dictionary produced by the ``to_dict`` method
        of this class.

        """
        if project is None:
            project = ws.new_project()

        # Uncategorize pore/throat and labels/properties, if present
        fd = FlatDict(dct, delimiter=delim)
        # If . is the delimiter, replace with | otherwise things break
        if delim == '.':
            delim = ' | '
            for key in list(fd.keys()):
                new_key = key.replace('.', delim)
                fd[new_key] = fd.pop(key)
        d = FlatDict(delimiter=delim)
        for key in list(fd.keys()):
            new_key = key.replace('pore' + delim, 'pore.')
            new_key = new_key.replace('throat' + delim, 'throat.')
            new_key = new_key.replace('labels' + delim, '')
            new_key = new_key.replace('properties' + delim, '')
            d[new_key] = fd.pop(key)

        # Plase data into correctly categorized dicts, for later handling
        objs = {'network': NestedDict(),
                'geometry': NestedDict(),
                'physics': NestedDict(),
                'phase': NestedDict(),
                'algorithm': NestedDict(),
                'base': NestedDict()}
        for item in d.keys():
            path = item.split(delim)
            if len(path) > 2:
                if path[-3] in objs.keys():
                    # Item is categorized by type, so note it
                    objs[path[-3]][path[-2]][path[-1]] = d[item]
                else:
                    # item is nested, not categorized; make it a base
                    objs['base'][path[-2]][path[-1]] = d[item]
            else:
                # If not categorized by type, make it a base
                objs['base'][path[-2]][path[-1]] = d[item]

        # Convert to OpenPNM Objects, attempting to infer type
        for objtype in objs.keys():
            for name in objs[objtype].keys():
                # Create empty object, using dummy name to avoid error
                obj = project._new_object(objtype=objtype, name='')
                # Overwrite name
                obj._set_name(name=name, validate=False)
                # Update new object with data from dict
                obj.update(objs[objtype][name])

        return project

    @classmethod
    def to_dict(cls, network=None, phases=[], element=['pore', 'throat'],
                interleave=True, flatten=True, categorize_by=[]):
        r"""
        Returns a single dictionary object containing data from the given
        OpenPNM objects, with the keys organized differently depending on
        optional arguments.

        Parameters
        ----------
        network : OpenPNM Network Object (optional)
            The network containing the desired data

        phases : list of OpenPNM Phase Objects (optional, default is none)
            A list of phase objects whose data are to be included

        element : string or list of strings
            An indication of whether 'pore' and/or 'throat' data are desired.
            The default is both.

        interleave : boolean (default is ``True``)
            When ``True`` (default) the data from all Geometry objects (and
            Physics objects if ``phases`` are given) is interleaved into
            a single array and stored as a network property (or Phase
            property for Physics data). When ``False``, the data for each
            object are stored under their own dictionary key, the structuring
            of which depends on the value of the ``flatten`` argument.

        flatten : boolean (default is ``True``)
            When ``True``, all objects are accessible from the top level
            of the dictionary.  When ``False`` objects are nested under their
            parent object.  If ``interleave`` is ``True`` this argument is
            ignored.

        categorize_by : string or list of strings
            Indicates how the dictionaries should be organized.  The list can
            contain any, all or none of the following strings:

            **'object'** : If specified the dictionary keys will be stored
            under a general level corresponding to their type (e.g.
            'network/net_01/pore.all'). If  ``interleave`` is ``True`` then
            only the only categories are *network* and *phase*, since
            *geometry* and *physics* data get stored under their respective
            *network* and *phase*.

            **'data'** : If specified the data arrays are additionally
            categorized by ``label`` and ``property`` to separate *boolean*
            from *numeric* data.

            **'element'** : If specified the data arrays are
            additionally categorized by ``pore`` and ``throat``, meaning
            that the propnames are no longer prepended by a 'pore.' or
            'throat.'

        Returns
        -------
        A dictionary with the data stored in a hierarchical data structure, the
        actual format of which depends on the arguments to the function.

        Notes
        -----
        There is a handy package called *flatdict* that can be used to
        access this dictionary using a single key such that:

        ``d[level_1][level_2] == d[level_1/level_2]``

        Importantly, converting to a *flatdict* allows it be converted to an
        *HDF5* file directly, since the hierarchy is dictated by the placement
        of '/' characters.
        """
        project, network, phases = cls._parse_args(network=network,
                                                   phases=phases)
        delim = ' | '
        d = NestedDict(delimiter=delim)

        def build_path(obj, key):
            propname = delim + key
            prefix = 'root'
            datatype = ''
            arr = obj[key]
            if 'object' in categorize_by:
                prefix = obj._isa()
            if 'element' in categorize_by:
                propname = delim + key.replace('.', delim)
            if 'data' in categorize_by:
                if arr.dtype == bool:
                    datatype = delim + 'labels'
                else:
                    datatype = delim + 'properties'
            path = prefix + delim + obj.name + datatype + propname
            return path

        for net in network:
            for key in net.keys(element=element, mode='all'):
                path = build_path(obj=net, key=key)
                d[path] = net[key]

            for geo in project.geometries().values():
                for key in geo.keys(element=element, mode='all'):
                    if interleave:
                        path = build_path(obj=net, key=key)
                        d[path] = net[key]
                    else:
                        path = build_path(obj=geo, key=key)
                        if flatten:
                            d[path] = geo[key]
                        elif 'object' in categorize_by:
                            path = path.split(delim)
                            path.insert(0, 'network')
                            path.insert(1, net.name)
                            path = delim.join(path)
                        else:
                            path = path.split(delim)
                            path.insert(1, net.name)
                            path = delim.join(path)
                        d[path] = geo[key]

        for phase in phases:
            for key in phase.keys(element=element, mode='all'):
                path = build_path(obj=phase, key=key)
                d[path] = phase[key]

            for phys in project.find_physics(phase=phase):
                if phys:
                    for key in phys.keys(element=element, mode='all'):
                        if interleave:
                            path = build_path(obj=phase, key=key)
                            d[path] = phase[key]
                        else:
                            path = build_path(obj=phys, key=key)
                            if flatten:
                                d[path] = phys[key]
                            elif 'object' in categorize_by:
                                path = path.split(delim)
                                path.insert(0, 'phase')
                                path.insert(1, phase.name)
                                path = delim.join(path)
                            else:
                                path = path.split(delim)
                                path.insert(1, phase.name)
                                path = delim.join(path)
                            d[path] = phys[key]

        if 'root' in d.keys():
            d = d['root']
        if 'project' in categorize_by:
            new_d = NestedDict()
            new_d[project.name] = d
            d = new_d
        return d

    @classmethod
    def save(cls, dct, filename):
        r"""
        Saves data from the given dictionary into the specified file.

        Parameters
        ----------
        dct : dictionary
            A dictionary to save to file, presumably obtained from the
            ``to_dict`` method of this class.

        filename : string or path object
            The filename to store the dictionary.

        """
        fname = cls._parse_filename(filename=filename, ext='dct')
        dct = sanitize_dict(dct)
        with open(fname, 'wb') as f:
            pickle.dump(dct, f)

    @classmethod
    def load(cls, filename):
        r"""
        Load data from the specified file into a Python dictionary

        Parameters
        ----------
        filename : string
            The path to the file to be opened

        Notes
        -----
        This returns a Python dictionary which can be converted into OpenPNM
        objects using the ``from_dict`` method of this class.

        """
        fname = cls._parse_filename(filename)
        with open(fname, 'rb') as f:
            dct = pickle.load(f)
        return dct



import time
import openpnm
import numpy as np
from copy import deepcopy
from openpnm.utils import SettingsDict, HealthDict, Workspace, logging
logger = logging.getLogger(__name__)
ws = Workspace()


class Project(list):
    r"""
    This class provides a container for all OpenPNM objects in a given
    simulation.

    A simulation is defined as a Network and all of it's associated objects.
    When instantiating a Network, a Project can be passed as an argument, but
    if not given one is created.  When instantiating any other object either
    a Network or a Project can be supplied.  In the former case, the
    Network's Project is retrieved and used.  The end result is that all
    objects are stored in a specific Project.

    The Project to which any object belongs can be retrieved with
    ``obj.project``.  Conversely, printing a Project displays a list of all
    objects it contains.

    Moreover, all Projects are registered with the Workspace.  Since there can
    be only instance of the Workspace it is possible to view all open Projects
    by printing the Workspace.

    See Also
    --------
    Workspace

    Notes
    -----
    The following table shows all the methods that are available on the Project
    objects along with a very brief description:

    +------------------+-------------------------------------------------+
    | Method           | Description                                     |
    +==================+=================================================+
    | export_data      | Export the pore and throat data from the giv... |
    +------------------+-------------------------------------------------+
    | import_data      |                                                 |
    +------------------+-------------------------------------------------+
    | purge_object     | Remove an object from the Project.  This rem... |
    +------------------+-------------------------------------------------+
    | load_object      | Loads a single object from a file               |
    +------------------+-------------------------------------------------+
    | save_object      | Saves the given object to a file                |
    +------------------+-------------------------------------------------+
    | find_physics     | Find the Physics object(s) associated with a... |
    +------------------+-------------------------------------------------+
    | find_geometry    | Find the Geometry associated with a given Ph... |
    +------------------+-------------------------------------------------+
    | find_phase       | Find the Phase associated with a given object.  |
    +------------------+-------------------------------------------------+
    | check_geometr... | Perform a check to find pores with overlappi... |
    +------------------+-------------------------------------------------+
    | check_physics... | Perform a check to find pores which have ove... |
    +------------------+-------------------------------------------------+

    """

    def __init__(self, *args, **kwargs):
        name = kwargs.pop('name', None)
        super().__init__(*args, **kwargs)
        # Register self with workspace
        ws[name] = self
        self.settings = SettingsDict()
        self.comments = 'Using OpenPNM ' + openpnm.__version__

    def extend(self, obj):
        r"""
        This function is used to add objects to the project.  Arguments can
        be single OpenPNM objects, an OpenPNM project list, or a plain list of
        OpenPNM objects.

        """
        if type(obj) is not list:
            obj = [obj]
        for item in obj:
            if hasattr(item, '_mro'):
                if 'GenericNetwork' in item._mro():
                    if self.network:
                        raise Exception('Project already has a network')
                # Must use append since extend breaks the dicts up into
                # separate objects, while append keeps it as a single object.
                super().append(item)
            else:
                raise Exception('Only OpenPNM objects can be added')

    def append(self, obj):
        r"""
        The Project (a list) must be kept as a flat list, so the append
        function, which can normally be used to insert a list into a list, is
        overloaded to basically prevent the normal append operation and simply
        calls ``extend``.

        """
        self.extend(obj)

    def remove(self, obj):
        r"""
        The given object is removed from the project

        This removes the object, along with all it's labels in associated
        objects, but does NOT remove the associated objects.

        See Also
        -------
        purge_object

        """
        self.purge_object(obj, deep=False)

    def pop(self, index):
        r"""
        The object at the given index is removed from the list and returned.

        Notes
        -----
        This method uses ``purge_object`` to perform the actual removal of the
        object. It is reommended to just use that directly instead.

        See Also
        --------
        purge_object

        """
        obj = self[index]
        self.purge_object(obj, deep=False)
        return obj

    def insert(self, index, obj):
        r"""
        Inserts the supplied object at the specified index in the Project list

        Notes
        -----
        The order of the objects in an OpenPNM Project lists do not matter, so
        it is recommended to just use ``append`` instead.

        See Also
        --------
        append
        extend

        """
        self.extend(obj)

    def clear(self, objtype=[]):
        r"""
        Clears objects from the project entirely or selectively, depdening on
        the received arguments.

        Parameters
        ----------
        objtype : list of strings
            A list containing the object type(s) to be removed.  If no types
            are specified, then all objects are removed.  To clear only objects
            of a specific type, use *'network'*, *'geometry'*, *'phase'*,
            *'physics'*, or *'algorithm'*.  It's also possible to use
            abbreviations, like *'geom'*.

        """
        if len(objtype) == 0:
            super().clear()
        else:
            names = [obj.name for obj in self]
            for name in names:
                try:
                    obj = self[name]
                    for t in objtype:
                        if obj._isa(t):
                            self.purge_object(obj)
                except KeyError:
                    pass

    def copy(self, name=None):
        r"""
        Creates a deep copy of the current project

        A deep copy means that new, unique versions of all the objects are
        created but with identical data and properties.

        Parameters
        ----------
        name : string
            The name to give to the new project.  If not supplied, a name
            is automatically generated.

        Returns
        -------
        proj : list
            A new Project object containing copies of all objects

        """
        if name is None:
            name = ws._gen_name()
        proj = deepcopy(self)
        ws[name] = proj
        return proj

    @property
    def workspace(self):
        return ws

    def _set_name(self, name):
        if name is None:
            name = ws._gen_name()
        ws[name] = self

    def _get_name(self):
        for key in ws.keys():
            if ws[key] is self:
                return key

    name = property(fget=_get_name, fset=_set_name)

    def __getitem__(self, key):
        if type(key) == str:
            obj = None
            for item in self:
                if item.name == key:
                    obj = item
            if obj is None:
                raise KeyError(key)
        else:
            obj = super().__getitem__(key)
        return obj

    def find_phase(self, obj):
        r"""
        Find the Phase associated with a given object.

        Parameters
        ----------
        obj : OpenPNM Object
            Can either be a Physics or Algorithm object

        Returns
        -------
        phase : OpenPNM Phase object

        Raises
        ------
        If no Phase object can be found, then an Exception is raised.
        """
        # If received phase, just return self
        if obj._isa('phase'):
            return obj
        # If phase happens to be in settings (i.e. algorithm), look it up
        if 'phase' in obj.settings.keys():
            phase = self.phases()[obj.settings['phase']]
            return phase
        # Otherwise find it using bottom-up approach (i.e. look in phase keys)
        for phase in self.phases().values():
            if ('pore.'+obj.name in phase) or ('throat.'+obj.name in phase):
                return phase
        # If all else fails, throw an exception
        raise Exception('Cannot find a phase associated with '+obj.name)

    def find_geometry(self, physics):
        r"""
        Find the Geometry associated with a given Physics

        Parameters
        ----------
        physics : OpenPNM Physics Object
            Must be a Physics object

        Returns
        -------
        geom : OpenPNM Geometry object

        Raises
        ------
        If no Geometry object can be found, then an Exception is raised.

        """
        # If geometry happens to be in settings, look it up directly
        if 'geometry' in physics.settings.keys():
            geom = self.geometries()[physics.settings['geometry']]
            return geom
        # Otherwise, use the bottom-up approach
        for geo in self.geometries().values():
            if physics in self.find_physics(geometry=geo):
                return geo
        # If all else fails, throw an exception
        raise Exception('Cannot find a geometry associated with '+physics.name)

    def find_physics(self, geometry=None, phase=None):
        r"""
        Find the Physics object(s) associated with a given Geometry, Phase,
        or combination.

        Parameters
        ----------
        geometry : OpenPNM Geometry Object
            The Geometry object for which the Physics object(s) are sought

        phase : OpenPNM Phase Object
            The Phase object for which the Physics object(s) are sought

        Returns
        -------
        physics : list
            A list containing the Physics object(s).  If only a ``geometry`` is
            specified the the Physics for all Phases is returned.  If only a
            ``phase`` is specified, then the Physics for all Geometries is
            returned.  If both ``geometry`` and ``phase`` is specified then
            the list only contains a single Physics.  If no Physics is found,
            the the list will be empty.  See the Notes section for more
            information.

        See Also
        --------
        grid

        Notes
        -----
        The Project has an ``grid`` attribute that shows the association of
        all objects.  If each Geometry represents a row and each Phase is a
        column, then each row/col intersection represents a Physics. This
        method finds the PHysics' at each intersection

        """

        if geometry and phase:
            physics = self.find_physics(geometry=geometry)
            phases = list(self.phases().values())
            phys = physics[phases.index(phase)]
            return phys
        elif geometry:
            result = []
            net = self.network
            geoPs = net['pore.'+geometry.name]
            geoTs = net['throat.'+geometry.name]
            for phase in self.phases().values():
                physics = self.find_physics(phase=phase)
                temp = None
                for phys in physics:
                    Ps = phase.map_pores(pores=phys.Ps, origin=phys)
                    physPs = phase.tomask(pores=Ps)
                    Ts = phase.map_throats(throats=phys.Ts, origin=phys)
                    physTs = phase.tomask(throats=Ts)
                    if np.all(geoPs == physPs) and np.all(geoTs == physTs):
                        temp = phys
                result.append(temp)
            return result
        elif phase:
            names = set(self.physics().keys())
            keys = set([item.split('.')[-1] for item in phase.keys()])
            hits = names.intersection(keys)
            phys = [self.physics().get(i, None) for i in hits]
            return phys
        else:
            phys = list(self.physics().values())
            return phys

    def find_full_domain(self, obj):
        r"""
        Find the full domain object associated with a given object.
        For geometry the network is found, for physics the phase is found and
        for all other objects which are defined for for the full domain,
        themselves are found.

        Parameters
        ----------
        obj : OpenPNM Object
            Can be any object

        Returns
        -------
        obj : An OpenPNM object

        """
        if 'Subdomain' not in obj._mro():
            # Network, Phase, Alg
            return obj
        else:
            if obj._isa() == 'geometry':
                # Geom
                return self.network
            else:
                # Phys
                return self.find_phase(obj)

    def _validate_name(self, name):
        if name in self.names:
            raise Exception('An object already exists named '+name)
        for item in self:
            for key in item.keys():
                if key.split('.')[1] == name:
                    raise Exception('A property/label is already named '+name)

    def _generate_name(self, obj):
        prefix = obj.settings['prefix']
        num = str(len([item for item in self if item._isa() == obj._isa()]))
        name = prefix + '_' + num.zfill(2)
        return name

    @property
    def names(self):
        names = [i.name for i in self]
        return names

    def purge_object(self, obj, deep=False):
        r"""
        Remove an object from the Project.  This removes all references to
        the object from all other objects (i.e. removes labels)

        Parameters
        ----------
        obj : OpenPNM Object or list of objects
            The object(s) to purge

        deep : boolean
            A flag that indicates whether to remove associated objects.
            If ``True``, then removing a Geometry or Phase also removes
            the associated Physics objects.  If ``False`` (default) then
            only the given object is removed, along with its labels in all
            associated objects.  Removing a Physics always keeps associated
            Geometry and Phases since they might also be associated with other
            Physics objects.

        Raises
        ------
        An Exception is raised if the object is a Network.

        Notes
        -----
        For a clearer picture of this logic, type ``print(project.grid)`` at
        the console.  A deep purge of a Geometry is like removing a row, while
        a Phase is like removing a column.

        """
        if type(obj) == list:
            for item in obj:
                self.purge_object(obj=item, deep=deep)
            return
        if obj._isa() in ['physics', 'algorithm']:
            self._purge(obj)
        if obj._isa() == 'geometry':
            if deep:
                physics = self.find_physics(geometry=obj)
                for phys in physics:
                    self._purge(self.physics()[phys.name])
            self._purge(obj)
        if obj._isa() == 'phase':
            if deep:
                physics = self.find_physics(phase=obj)
                for phys in physics:
                    self._purge(self.physics()[phys.name])
            self._purge(obj)
        if obj._isa() == 'network':
            raise Exception('Cannot purge a network, just make a new project')

    def _purge(self, obj):
        for item in self:
            for key in list(item.keys()):
                if key.split('.')[-1] == obj.name:
                    del item[key]
        super().remove(obj)

    def save_object(self, obj):
        r"""
        Saves the given object or list of objects to a file

        Parameters
        ----------
        obj : OpenPNM object or list of objects
            The objects to be saved.  Depending on the object type, the file
            extension will be one of 'net', 'geo', 'phase', 'phys' or 'alg'.
        """
        from openpnm.io import OpenpnmIO
        OpenpnmIO.save_object_to_file(objs=obj)

    def load_object(self, filename):
        r"""
        Loads a single object from a file

        Parameters
        ----------
        filename : string or path object
            The name of the file containing the saved object.  Can include
            an absolute or relative path as well.  If only a filename is
            given it will be saved in the current working directory.  The
            object type is inferred from

        """
        from openpnm.io import OpenpnmIO
        OpenpnmIO.load_object_from_file(filename=filename, project=self)

    def save_project(self, filename=''):
        r"""
        Save the current project to a ``pnm`` file.

        Parameters
        ----------
        filename : string or path object
            The name of the file.  Can include an absolute or relative path
            as well.  If only a filename is given it will be saved in the
            current working directory.

        """
        ws.save_project(project=self, filename=filename)

    def _new_object(self, objtype, name=None):
        r"""
        """
        if objtype.startswith('net'):
            obj = openpnm.network.GenericNetwork(project=self, name=name)
        elif objtype.startswith('geo'):
            obj = openpnm.geometry.GenericGeometry(project=self, name=name,
                                                   pores=[], throats=[])
        elif objtype.startswith('pha'):
            obj = openpnm.phases.GenericPhase(project=self, name=name)
        elif objtype.startswith('phy'):
            obj = openpnm.physics.GenericPhysics(project=self, name=name)
        elif objtype.startswith('alg'):
            obj = openpnm.algorithm.GenericAlgorithm(project=self, name=name)
        else:
            obj = openpnm.core.Base(project=self, name=name)
        return obj

    def import_data(self, filename):
        r"""
        """
        raise NotImplementedError('Use the io module to import data')

    def export_data(self, phases=[], filename=None, filetype='vtp'):
        r"""
        Export the pore and throat data from the given object(s) into the
        specified file and format.

        Parameters
        ----------
        phases : list of OpenPNM Phase Objects
            The data on each supplied phase will be added to file

        filename : string
            The file name to use.  If none is supplied then one will be
            automatically generated based on the name of the project
            containing the supplied Network, with the date and time appended.

        filetype : string
            Which file format to store the data.  If a valid extension is
            included in the ``filename``, this is ignored.  Option are:

            **'vtk'** : (default) The Visualization Toolkit format, used by
            various softwares such as Paraview.  This actually produces a 'vtp'
            file.  NOTE: This can be quite slow since all the data is written
            to a simple text file.  For large data simulations consider
            'xdmf'.

            **'csv'** : The comma-separated values format, which is easily
            openned in any spreadsheet program.  The column names represent
            the property name, including the type and name of the object to
            which they belonged, all separated by the pipe character.

            **'xmf'** : The extensible data markup format, is a very efficient
            format for large data sets.  This actually results in the creation
            of two files, the *xmf* file and an associated *hdf* file.  The
            *xmf* file contains instructions for looking into the *hdf* file
            where the data is stored. Paraview opens the *xmf* format natively,
            and is very fast.

            **'mat'** : Matlab 'mat-file', which can be openned in Matlab.

        Notes
        -----
        This is a helper function for the actual functions in the IO module.
        For more control over the format of the output, and more information
        about the format refer to ``openpnm.io``.

        """
        project = self
        network = self.network
        if filename is None:
            filename = project.name + '_' + time.strftime('%Y%b%d_%H%M%p')
        # Infer filetype from extension on file name...if given
        if '.' in filename:
            exts = ['vtk', 'vtp', 'vtu', 'csv', 'xmf', 'xdmf', 'hdf', 'hdf5',
                    'h5', 'mat']
            if filename.split('.')[-1] in exts:
                filename, filetype = filename.rsplit('.', 1)
        if filetype.lower() in ['vtk', 'vtp', 'vtu']:
            openpnm.io.VTK.save(network=network, phases=phases,
                                filename=filename)
        if filetype.lower() == 'csv':
            openpnm.io.CSV.save(network=network, phases=phases,
                                filename=filename)
        if filetype.lower() in ['xmf', 'xdmf']:
            openpnm.io.XDMF.save(network=network, phases=phases,
                                 filename=filename)
        if filetype.lower() in ['hdf5', 'hdf', 'h5']:
            f = openpnm.io.HDF5.to_hdf5(network=network, phases=phases,
                                        filename=filename)
            f.close()
        if filetype.lower() == 'mat':
            openpnm.io.MAT.save(network=network, phases=phases,
                                filename=filename)
        if filetype == 'COMSOL':
            openpnm.io.COMSOL.save(network=network, phases=phases,
                                   filename=filename)

    def _dump_data(self, mode=['props']):
        r"""
        Dump data from all objects in project to an HDF5 file.  Note that
        'pore.coords', 'throat.conns', 'pore.all', 'throat.all', and all
        labels pertaining to the linking of objects are kept.

        Parameters
        ----------
        mode : string or list of strings
            The type of data to be dumped to the HDF5 file.  Options are:

            **'props'** : Numerical data such as 'pore.diameter'.  The default
            is only 'props'.

            **'labels'** : Boolean data that are used as labels.  Since this
            is boolean data it does not consume large amounts of memory and
            probably does not need to be dumped.

        See Also
        --------
        _fetch_data

        Notes
        -----
        In principle, after data is fetched from an HDF5 file, it should
        physically stay there until it's called upon.  This let users manage
        the data as if it's in memory, even though it isn't.  This behavior
        has not been confirmed yet, which is why these functions are hidden.

        """
        import h5py
        with h5py.File(self.name + '.hdf5') as f:
            for obj in self:
                for key in list(obj.keys()):
                    tempname = obj.name + '|' + '_'.join(key.split('.'))
                    arr = obj[key]
                    if 'U' in str(obj[key][0].dtype):
                        pass
                    elif 'all' in key.split('.'):
                        pass
                    else:
                        f.create_dataset(name='/'+tempname, shape=arr.shape,
                                         dtype=arr.dtype, data=arr)
            for obj in self:
                obj.clear(mode=mode)

    def _fetch_data(self):
        r"""
        Retrieve data from an HDF5 file and place onto correct objects in the
        project

        See Also
        --------
        _dump_data

        Notes
        -----
        In principle, after data is fetched from and HDF5 file, it should
        physically stay there until it's called upon.  This let users manage
        the data as if it's in memory, even though it isn't.  This behavior
        has not been confirmed yet, which is why these functions are hidden.

        """
        import h5py
        with h5py.File(self.name + '.hdf5') as f:
            # Reload data into project
            for item in f.keys():
                obj_name, propname = item.split('|')
                propname = propname.split('_')
                propname = propname[0] + '.' + '_'.join(propname[1:])
                self[obj_name][propname] = f[item]

    @property
    def network(self):
        net = list(self._get_objects_by_type('network').values())
        if len(net) > 0:
            net = net[0]
        else:
            net = None
        return net

    def geometries(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('geometry')

    def phases(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('phase')

    def physics(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('physics')

    def algorithms(self, name=None):
        if name:
            return self._get_object_by_name(name)
        else:
            return self._get_objects_by_type('algorithm')

    def _get_object_by_name(self, name):
        for item in self:
            if item.name == name:
                return item
        raise Exception('An object named ' + name + ' was not found')

    def _get_objects_by_type(self, objtype):
        return {item.name: item for item in self if item._isa(objtype)}

    def _set_comments(self, string):
        if hasattr(self, '_comments') is False:
            self._comments = {}
        self._comments[time.strftime('%c')] = string

    def _get_comments(self):
        if hasattr(self, '_comments') is False:
            self._comments = {}
        for key in list(self._comments.keys()):
            print(key, ': ', self._comments[key])

    comments = property(fget=_get_comments, fset=_set_comments)

    def __str__(self):
        s = []
        hr = '―'*78
        s.append(hr)
        s.append(' {0:<15} '.format('Object Name')
                 + '{0:<65}'.format('Object ID'))
        s.append(hr)
        for item in self:
            s.append(' {0:<15} '.format(item.name)
                     + '{0:<65}'.format(item.__repr__()))
        s.append(hr)
        return '\n'.join(s)

    def check_geometry_health(self):
        r"""
        Perform a check to find pores with overlapping or undefined Geometries

        Returns
        -------
        A HealthDict
        """
        health = HealthDict()
        health['overlapping_pores'] = []
        health['undefined_pores'] = []
        health['overlapping_throats'] = []
        health['undefined_throats'] = []
        geoms = self.geometries().keys()
        if len(geoms):
            net = self.network
            Ptemp = np.zeros((net.Np,))
            Ttemp = np.zeros((net.Nt,))
            for item in geoms:
                Pind = net['pore.'+item]
                Tind = net['throat.'+item]
                Ptemp[Pind] = Ptemp[Pind] + 1
                Ttemp[Tind] = Ttemp[Tind] + 1
            health['overlapping_pores'] = np.where(Ptemp > 1)[0].tolist()
            health['undefined_pores'] = np.where(Ptemp == 0)[0].tolist()
            health['overlapping_throats'] = np.where(Ttemp > 1)[0].tolist()
            health['undefined_throats'] = np.where(Ttemp == 0)[0].tolist()
        else:
            health['undefined_pores'] = self.network.Ps
            health['undefined_throats'] = self.network.Ts
        return health

    def check_physics_health(self, phase):
        r"""
        Perform a check to find pores which have overlapping or missing Physics

        Parameters
        ----------
        phase : OpenPNM Phase object
            The Phase whose Physics should be checked

        Returns
        -------
        A HealthDict

        """
        health = HealthDict()
        health['overlapping_pores'] = []
        health['undefined_pores'] = []
        health['overlapping_throats'] = []
        health['undefined_throats'] = []
        geoms = self.geometries().keys()
        if len(geoms):
            phys = self.find_physics(phase=phase)
            if len(phys) == 0:
                raise Exception(str(len(geoms))+' geometries were found, but'
                                + ' no physics')
            if None in phys:
                raise Exception('Undefined physics found, check the grid')
            Ptemp = np.zeros((phase.Np,))
            Ttemp = np.zeros((phase.Nt,))
            for item in phys:
                Pind = phase['pore.'+item.name]
                Tind = phase['throat.'+item.name]
                Ptemp[Pind] = Ptemp[Pind] + 1
                Ttemp[Tind] = Ttemp[Tind] + 1
            health['overlapping_pores'] = np.where(Ptemp > 1)[0].tolist()
            health['undefined_pores'] = np.where(Ptemp == 0)[0].tolist()
            health['overlapping_throats'] = np.where(Ttemp > 1)[0].tolist()
            health['undefined_throats'] = np.where(Ttemp == 0)[0].tolist()
        return health

    def check_data_health(self, obj):
        r"""
        Check the health of pore and throat data arrays.

        Parameters
        ----------
        obj : OpenPNM object
            A handle of the object to be checked

        Returns
        -------
        health : dict
            Returns a HealthDict object which a basic dictionary with an added
            ``health`` attribute that is True is all entries in the dict are
            deemed healthy (empty lists), or False otherwise.

        """
        health = HealthDict()
        for item in obj.props():
            health[item] = []
            if obj[item].dtype == 'O':
                health[item] = 'No checks on object'
            elif np.sum(np.isnan(obj[item])) > 0:
                health[item] = 'Has NaNs'
            elif np.shape(obj[item])[0] != obj._count(item.split('.')[0]):
                health[item] = 'Wrong Length'
        return health

    def check_network_health(self):
        r"""
        This method check the network topological health by checking for:

            (1) Isolated pores
            (2) Islands or isolated clusters of pores
            (3) Duplicate throats
            (4) Bidirectional throats (ie. symmetrical adjacency matrix)
            (5) Headless throats

        Returns
        -------
        health : dict
            A dictionary containing the offending pores or throat numbers under
            each named key.

        Notes
        -----
        It also returns a list of which pores and throats should be trimmed
        from the network to restore health.  This list is a suggestion only,
        and is based on keeping the largest cluster and trimming the others.

        Notes
        -----
        - Does not yet check for duplicate pores
        - Does not yet suggest which throats to remove
        - This is just a 'check' and does not 'fix' the problems it finds
        """
        import scipy.sparse.csgraph as csg
        import scipy.sparse as sprs

        health = HealthDict()
        health['disconnected_clusters'] = []
        health['isolated_pores'] = []
        health['trim_pores'] = []
        health['duplicate_throats'] = []
        health['bidirectional_throats'] = []
        health['headless_throats'] = []
        health['looped_throats'] = []

        net = self.network

        # Check for headless throats
        hits = np.where(net['throat.conns'] > net.Np - 1)[0]
        if np.size(hits) > 0:
            health['headless_throats'] = np.unique(hits)
            return health

        # Check for throats that loop back onto the same pore
        P12 = net['throat.conns']
        hits = np.where(P12[:, 0] == P12[:, 1])[0]
        if np.size(hits) > 0:
            health['looped_throats'] = hits

        # Check for individual isolated pores
        Ps = net.num_neighbors(net.pores())
        if np.sum(Ps == 0) > 0:
            health['isolated_pores'] = np.where(Ps == 0)[0]

        # Check for separated clusters of pores
        temp = []
        am = net.create_adjacency_matrix(fmt='coo', triu=True)
        Cs = csg.connected_components(am, directed=False)[1]
        if np.unique(Cs).size > 1:
            for i in np.unique(Cs):
                temp.append(np.where(Cs == i)[0])
            b = np.array([len(item) for item in temp])
            c = np.argsort(b)[::-1]
            for i in range(0, len(c)):
                health['disconnected_clusters'].append(temp[c[i]])
                if i > 0:
                    health['trim_pores'].extend(temp[c[i]])

        # Check for duplicate throats
        am = net.create_adjacency_matrix(fmt='csr', triu=True).tocoo()
        hits = np.where(am.data > 1)[0]
        if len(hits):
            mergeTs = []
            hits = np.vstack((am.row[hits], am.col[hits])).T
            ihits = hits[:, 0] + 1j*hits[:, 1]
            conns = net['throat.conns']
            iconns = conns[:, 0] + 1j*conns[:, 1]  # Convert to imaginary
            for item in ihits:
                mergeTs.append(np.where(iconns == item)[0])
            health['duplicate_throats'] = mergeTs

        # Check for bidirectional throats
        adjmat = net.create_adjacency_matrix(fmt='coo')
        num_full = adjmat.sum()
        temp = sprs.triu(adjmat, k=1)
        num_upper = temp.sum()
        if num_full > num_upper:
            biTs = np.where(net['throat.conns'][:, 0]
                            > net['throat.conns'][:, 1])[0]
            health['bidirectional_throats'] = biTs.tolist()

        return health

    def inspect_locations(self, element, indices, objs=[], mode='all'):
        r"""
        Shows the values of all props and/or labels for a given subset of
        pores or throats.

        Parameters
        ----------
        element : str
            The type of locations to inspect, either 'pores', or 'throats'
        indices : array_like
            The pore or throat indices to inspect
        objs : list of OpenPNM Objects
            If given, then only the properties on the recieved object are
            inspected.  If not given, then all objects are inspected (default).
        mode : list of strings
            Indicates whether to inspect 'props', 'labels', or 'all'.  The
            default is all

        Returns
        -------
        df : Pandas DataFrame
            A data frame object with each location as a column and each row
            as a property and/or label.
        """
        from pandas import DataFrame
        props = {}
        if type(objs) is not list:
            objs = [objs]
        if not objs:
            objs = self
        for obj in objs:
            d = {k: obj[k][indices] for k in obj.keys(element=element, mode=mode)}
            for item in list(d.keys()):
                if d[item].ndim > 1:
                    d.pop(item)
                    if item == 'pore.coords':
                        d['pore.coords_X'], d['pore.coords_Y'], \
                            d['pore.coords_Z'] = obj['pore.coords'][indices].T
                    if item == 'throat.conns':
                        d['throat.conns_head'], d['throat.conns_tail'] = \
                            obj['throat.conns'][indices].T
                _ = [props.update({obj.name+'.'+item: d[item]}) for item in d.keys()]
        df = DataFrame(props)
        df = df.rename(index={k: indices[k] for k in range(len(indices))})
        return df.T

    def _regenerate_models(self, objs=[], propnames=[]):
        r"""
        Can be used to regenerate models across all objects in the project.

        Parameters
        ----------
        objs : list of OpenPNM objects
            Can be used to specify which specific objects to regenerate.  The
            default is to regenerate all objects.  If a subset of objects is
            given, this function ensure they are generated in a sensible order
            such as any phases are done before any physics objects.

        propnames : list of strings, or string
            The specific model to regenerate.  If none are given then ALL
            models on all objects are regenerated.  If a subset is given,
            then only object that have a corresponding model are regenerated,
            to avoid any problems.  This means that a single model can be
            given, without specifying the objects.

        """
        objs = list(objs)
        if objs == []:
            objs = self
        if type(propnames) is str:
            propnames = [propnames]
        # Sort objs in the correct order (geom, phase, phys)
        net = [i for i in objs if i is self.network]
        geoms = [i for i in objs if i in self.geometries().values()]
        phases = [i for i in objs if i in self.phases().values()]
        phys = [i for i in objs if i in self.physics().values()]
        objs = net + geoms + phases + phys
        for obj in objs:
            if len(propnames):
                for model in propnames:
                    if model in obj.models.keys():
                        obj.regenerate_models(propnames=model)
            else:
                obj.regenerate_models()

    def get_grid(self, astype='table'):
        r"""
        """
        from pandas import DataFrame as df

        geoms = self.geometries().keys()
        phases = [p.name for p in self.phases().values() if not hasattr(p, 'mixture')]
        grid = df(index=geoms, columns=phases)
        for r in grid.index:
            for c in grid.columns:
                phys = self.find_physics(phase=self[c], geometry=self[r])
                if phys is not None:
                    grid.loc[r][c] = phys.name
                else:
                    grid.loc[r][c] = '---'
        if astype == 'pandas':
            pass
        elif astype == 'dict':
            grid = grid.to_dict()
        elif astype == 'table':
            from terminaltables import AsciiTable
            headings = [self.network.name] + list(grid.keys())
            g = [headings]
            for row in list(grid.index):
                g.append([row] + list(grid.loc[row]))
            grid = AsciiTable(g)
            grid.title = 'Project: ' + self.name
            grid.padding_left = 3
            grid.padding_right = 3
            grid.justify_columns = {col: 'center' for col in range(len(headings))}
        elif astype == 'grid':
            grid = ProjectGrid()
        return grid

    @property
    def grid(self):
        grid = self.get_grid(astype='table')
        obj = ProjectGrid(grid)
        return obj


class Grid():

    def __init__(self, table):
        self.grid = table

    def index(self):
        index = [row[0] for row in self.grid.table_data[1:]]
        return index

    def header(self):
        columns = self.grid.table_data[0][1:]
        return columns

    def row(self, name):
        for row in self.grid.table_data:
            if row[0].startswith(name):
                return row
        else:
            raise ValueError(name + 'is not in list')

    def col(self, name):
        # Find column number
        index = self.grid.table_data[0].index(name)
        col = []
        for row in self.grid.table_data:
            col.append(row[index])
        return col

    def __str__(self):
        s = self.grid.table.__str__()
        return s


class ProjectGrid(Grid):
    r"""
    This is a subclass of Grid, which adds the ability to lookup by geometries
    and phases, as more specific versions of rows and cols
    """

    def geometries(self):
        return self.index()

    def phases(self):
        return self.header()



import numpy as _np
from openpnm.io import GenericIO
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class STL(GenericIO):
    r"""
    The STL Format is a Standard Triangle (or Tessellation) Language supported
    by many CAD packages and used for 3D printing. Export to this format may be
    slow since a 3D closed surface mesh is built.
    Requires installation of "netgen".

    """

    @classmethod
    def save(cls, network, phases=[], filename='', maxsize='auto',
             fileformat='STL Format', logger_level=0):
        r"""
        Saves (transient/steady-state) data from the given objects into the
        specified file.

        Parameters
        ----------
        network : OpenPNM Network Object.
            The network containing the desired data.

        phases : list of OpenPNM Phase Objects (place holder, default is none).

        filename : string (optional).
            The name of the file containing the data to export.

        maxsize : a float or a string "auto" (optional).
            The maximum size of the mesh elements allowed. "auto" corresponds
            to an automatic determination based on pores and throats sizes. Any
            float value will be used as a maximum size. Small values result in
            finner meshes, but slower mesh calculations.

        fileformat : string (optional).
            Default is "STL Format" which corresponds to STL format. Other
            formats such as Gmsh and Fluent are supported (see ngsolve.org).

        logger_level : integer between 0 and 7 (optional).
            Default is 0. The logger level set in netgen package.

        Notes
        -----
        This method only saves the geometry of the network, not any of the
        pore-scale models or other attributes.  To save an actual OpenPNM
        Project use the ``Workspace`` object.

        """
        try:
            import netgen.csg as csg
        except ModuleNotFoundError:
            logger.error('Module "netgen.csg" not found.')
        try:
            from netgen.meshing import SetMessageImportance as log
            log(logger_level)
        except ModuleNotFoundError:
            logger.warning('Module "netgen.meshing" not found. '
                           + 'The "logger_level" ignored.')

        project, network, phases = cls._parse_args(network=network,
                                                   phases=phases)
        network = network[0]

        if filename == '':
            filename = project.name
        path = cls._parse_filename(filename=filename, ext='stl')
        # Path is a pathlib object, so slice it up as needed
        fname_stl = path.name

        # correct connections where 'pore.diameter' = 'throat.diameter'
        dt = network['throat.diameter'].copy()
        dp = network['pore.diameter'][network['throat.conns']]
        dt[dp[:, 0] == dt] *= 0.99
        dt[dp[:, 1] == dt] *= 0.99

        scale = max(network['pore.diameter'].max(), dt.max(),
                    network['throat.length'].max())
        if maxsize == 'auto':
            maxsize = min(network['pore.diameter'].min(), dt.min(),
                          network['throat.length'].min())
        geo = csg.CSGeometry()

        # define pores
        geometry = csg.Sphere(csg.Pnt(network['pore.coords'][0, 0]/scale,
                                      network['pore.coords'][0, 1]/scale,
                                      network['pore.coords'][0, 2]/scale),
                              network['pore.diameter'][0]/scale/2)
        for p in range(1, network.Np):
            pore = csg.Sphere(csg.Pnt(network['pore.coords'][p, 0]/scale,
                                      network['pore.coords'][p, 1]/scale,
                                      network['pore.coords'][p, 2]/scale),
                              network['pore.diameter'][p]/scale/2)
            geometry += pore

        # define throats
        for t in range(network.Nt):
            A = network['throat.endpoints.tail'][t, :]/scale
            B = network['throat.endpoints.head'][t, :]/scale
            V = (B-A)/_np.linalg.norm(B-A)
            plane1 = csg.Plane(csg.Pnt(A[0], A[1], A[2]),
                               csg.Vec(-V[0], -V[1], -V[2]))
            plane2 = csg.Plane(csg.Pnt(B[0], B[1], B[2]),
                               csg.Vec(V[0], V[1], V[2]))
            cylinder = csg.Cylinder(csg.Pnt(A[0], A[1], A[2]),
                                    csg.Pnt(B[0], B[1], B[2]),
                                    dt[t]/scale/2)
            throat = cylinder * plane1 * plane2
            geometry += throat

        # add pore and throats to geometry, build mesh, rescale, and export
        geo.Add(geometry)
        mesh = geo.GenerateMesh(maxh=maxsize/scale)
        mesh.Scale(scale)
        mesh.Export(filename=fname_stl, format=fileformat)



r"""

**openpnm.models.geometry**

----

This submodule contains pore-scale models that calculate geometrical properties

"""

from . import pore_size
from . import pore_seed
from . import pore_volume
from . import pore_surface_area
from . import pore_area
from . import throat_endpoints
from . import throat_area
from . import throat_seed
from . import throat_size
from . import throat_length
from . import throat_perimeter
from . import throat_surface_area
from . import throat_volume
from . import throat_shape_factor
from . import throat_centroid
from . import throat_vector



import os as os
import numpy as np
import scipy as sp
from pathlib import Path
from openpnm.utils import logging, Project
from openpnm.network import GenericNetwork
from openpnm.io import GenericIO
from openpnm.topotools import trim
logger = logging.getLogger(__name__)


class MARock(GenericIO):
    r"""
    3DMA-Rock is a network extraction algorithm developed by Brent Lindquist
    and his group

    It uses Medial Axis thinning to find the skeleton of the pore space, then
    extracts geometrical features such as pore volume and throat
    cross-sectional area.

    [1] Lindquist, W. Brent, S. M. Lee, W. Oh, A. B. Venkatarangan, H. Shin,
    and M. Prodanovic. "3DMA-Rock: A software package for automated analysis
    of rock pore structure in 3-D computed microtomography images." SUNY Stony
    Brook (2005).
    """

    @classmethod
    def load(cls, path, voxel_size=1, project=None):
        r"""
        Load data from a 3DMA-Rock extracted network.  This format consists of
        two files: 'rockname.np2th' and 'rockname.th2pn'.  They should be
        stored together in a folder which is referred to by the path argument.
        These files are binary and therefore not human readable.

        Parameters
        ----------
        path : string
            The location of the 'np2th' and 'th2np' files. This can be an
            absolute path or relative to the current working directory.

        network : OpenPNM Network Object
            If an Network object is recieved, this method will add new data to
            it but NOT overwrite anything that already exists.  This can be
            used to append data from different sources.

        voxel_size : scalar
            The resolution of the image on which 3DMA-Rock was run, in terms of
            the linear length of eac voxel. The default is 1.  This is used to
            scale the voxel counts to actual dimension. It is recommended that
            this value be in SI units [m] to work well with OpenPNM.

        project : OpenPNM Project object
            A GenericNetwork is created and added to the specified Project.
            If no Project is supplied then one will be created and returned.

        """

        net = {}
        path = Path(path)
        path = path.resolve()
        for file in os.listdir(path):
            if file.endswith(".np2th"):
                np2th_file = os.path.join(path, file)
            elif file.endswith(".th2np"):
                th2np_file = os.path.join(path, file)

        with open(np2th_file, mode='rb') as f:
            [Np, Nt] = np.fromfile(file=f, count=2, dtype='u4')
            net['pore.boundary_type'] = sp.ndarray([Np, ], int)
            net['throat.conns'] = np.ones([Nt, 2], int)*(-1)
            net['pore.coordination'] = sp.ndarray([Np, ], int)
            net['pore.ID_number'] = sp.ndarray([Np, ], int)
            for i in range(0, Np):
                ID = np.fromfile(file=f, count=1, dtype='u4')
                net['pore.ID_number'][i] = ID
                net['pore.boundary_type'][i] = np.fromfile(file=f, count=1,
                                                           dtype='u1')
                z = np.fromfile(file=f, count=1, dtype='u4')[0]
                net['pore.coordination'][i] = z
                att_pores = np.fromfile(file=f, count=z, dtype='u4')
                att_throats = np.fromfile(file=f, count=z, dtype='u4')
                for j in range(0, len(att_throats)):
                    t = att_throats[j] - 1
                    p = att_pores[j] - 1
                    net['throat.conns'][t] = [i, p]
            net['throat.conns'] = np.sort(net['throat.conns'], axis=1)
            net['pore.volume'] = np.fromfile(file=f, count=Np, dtype='u4')
            nx = np.fromfile(file=f, count=1, dtype='u4')
            nxy = np.fromfile(file=f, count=1, dtype='u4')
            pos = np.fromfile(file=f, count=Np, dtype='u4')
            ny = nxy/nx
            ni = np.mod(pos, nx)
            nj = np.mod(np.floor(pos/nx), ny)
            nk = np.floor(np.floor(pos/nx)/ny)
            net['pore.coords'] = np.array([ni, nj, nk]).T

        with open(th2np_file, mode='rb') as f:
            Nt = np.fromfile(file=f, count=1, dtype='u4')[0]
            net['throat.area'] = np.ones([Nt, ], dtype=int)*(-1)
            for i in range(0, Nt):
                ID = np.fromfile(file=f, count=1, dtype='u4')
                net['throat.area'][i] = np.fromfile(file=f, count=1,
                                                    dtype='f4')
                # numvox = np.fromfile(file=f, count=1, dtype='u4')
                att_pores = np.fromfile(file=f, count=2, dtype='u4')
            nx = np.fromfile(file=f, count=1, dtype='u4')
            nxy = np.fromfile(file=f, count=1, dtype='u4')
            pos = np.fromfile(file=f, count=Nt, dtype='u4')
            ny = nxy/nx
            ni = np.mod(pos, nx)
            nj = np.mod(np.floor(pos/nx), ny)
            nk = np.floor(np.floor(pos/nx)/ny)
            net['throat.coords'] = np.array([ni, nj, nk]).T
            net['pore.internal'] = net['pore.boundary_type'] == 0

        # Convert voxel area and volume to actual dimensions
        net['throat.area'] = (voxel_size**2)*net['throat.area']
        net['pore.volume'] = (voxel_size**3)*net['pore.volume']

        if project is None:
            project = Project(name=path)
        network = GenericNetwork(project=project)
        network = cls._update_network(network=network, net=net)

        # Trim headless throats before returning
        ind = np.where(network['throat.conns'][:, 0] == -1)[0]
        trim(network=network, throats=ind)

        return project



r"""

.. autofunction:: openpnm.models.physics.multiphase.conduit_conductance
.. autofunction:: openpnm.models.physics.multiphase.late_filling

"""
import numpy as np
import scipy as sp


def conduit_conductance(target, throat_conductance,
                        throat_occupancy='throat.occupancy',
                        pore_occupancy='pore.occupancy',
                        mode='strict', factor=1e-6):
    r"""
    Determines the conductance of a pore-throat-pore conduit based on the
    invaded state of each element.

    Parameters
    ----------
    target : OpenPNM Object
        The OpenPNM object where the model is attached.  Should either be a
        Physics or a Phase.

    throat_conductance : string
        The transport conductance of the phase associated with the ``target``
        object at single-phase conditions.

    pore_occupancy : string
        The property name containing the occupancy of the phase associated
        with the ``target`` object.  An occupancy of 1 means the pore
        is completely filled with the phase and it fully conducts.

    throat_occupancy : string
        The property name containing the occupancy of the phase associated
        with the ``target`` object.  An occupancy of 1 means the throat
        is completely filled with the phase and it fully conducts.

    mode : 'strict' or 'medium' or 'loose'
        How agressive the method should be when determining if a conduit is
        closed.

        **'strict'** :  If any pore or throat in the conduit is unoccupied by
         the given phase, the conduit is closed.

        **'medium'** : If either the throat or both pores are unoccupied, the
        conduit is closed

        **'loose'** : Only close the conduit if the throat is unoccupied

    factor : float (default is 1e-6)
        The factor which becomes multiplied to the original conduit's
        conductance to severely limit transport, but not set it to zero.

    Returns
    -------
    value : NumPy ndarray
        Array containing conduit conductance values.

    """
    network = target.project.network
    phase = target.project.find_phase(target)
    Tinv = phase[throat_occupancy] < 0.5
    P12 = network['throat.conns']
    Pinv = phase[pore_occupancy][P12] < 0.5
    if mode == 'loose':
        mask = Tinv
    elif mode == 'medium':
        mask = Tinv + np.all(Pinv, axis=1)
    elif mode == 'strict':
        mask = Tinv + np.any(Pinv, axis=1)
    else:
        raise Exception('Unrecongnized mode '+mode)
    value = phase[throat_conductance].copy()
    value[mask] = value[mask]*factor
    # Now map throats onto target object
    Ts = network.map_throats(throats=target.Ts, origin=target)
    return value[Ts]


def late_filling(target, pressure='pore.pressure',
                 Pc_star='pore.pc_star',
                 Swp_star=0.2, eta=3):
    r"""
    Calculates the fraction of a pore or throat filled with invading fluid
    based on the capillary pressure in the invading phase.  The invading phase
    volume is calculated from:

        .. math::
            S_{nwp} = 1 - S_{wp}^{*} (P^{*}/P_{c})^{\eta}

    Parameters
    ----------
    pressure : string
        The capillary pressure in the non-wetting phase (Pc > 0).

    Pc_star : string
        The minimum pressure required to create an interface within the pore
        body or throat.  Typically this would be calculated using the Washburn
        equation.

    Swp_star : float
        The residual wetting phase in an invaded pore or throat at a pressure
        of ``pc_star``.

    eta : float
        Exponent controlling the rate at which wetting phase is displaced with
        increasing pressure.

    Returns
    -------
    An array containing the fraction of each pore or throat that would be
    filled with non-wetting phase at the given phase pressure.  This does not
    account for whether or not the element is actually invaded, which requires
    a percolation algorithm of some sort.

    """
    element = pressure.split('.')[0]
    network = target.project.network
    phase = target.project.find_phase(target)
    pc_star = phase[Pc_star]
    Pc = phase[pressure]
    # Remove any 0's from the Pc array to prevent numpy div by 0 warning
    Pc = np.maximum(Pc, 1e-9)
    Swp = Swp_star*((pc_star/Pc)**eta)
    values = np.clip(1 - Swp, 0.0, 1.0)
    # Now map element onto target object
    if element == 'throat':
        Ts = network.map_throats(throats=target.Ts, origin=target)
        values = values[Ts]
    else:
        Ps = network.map_pores(pores=target.Ps, origin=target)
        values = values[Ps]
    return values



r"""

.. autofunction:: openpnm.models.geometry.throat_shape_factor.compactness
.. autofunction:: openpnm.models.geometry.throat_shape_factor.mason_morrow
.. autofunction:: openpnm.models.geometry.throat_shape_factor.jenkins_rao

"""
import numpy as _np
import scipy as _sp


def compactness(target, throat_perimeter='throat.perimeter',
                throat_area='throat.area'):
    r"""
    Mortensen et al. have shown that the Hagen-Poiseuille hydraluic resistance
    is linearly dependent on the compactness. Defined as perimeter^2/area.
    The dependence is not universal as shapes with sharp corners provide more
    resistance than those that are more elliptical. Count the number of
    vertices and apply the right correction.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_perimeter : string
        The dictionary key of the array containing the throat perimeter values.

    throat_area : string
        The dictionary key of the array containing the throat area values.

    Returns
    -------
    alpha : NumPy ndarray
        Array containing throat compactness values.

    References
    ----------
    Mortensen N.A, Okkels F., and Bruus H. Reexamination of Hagen-Poiseuille
    flow: Shape dependence of the hydraulic resistance in microchannels.
    Physical Review E, v.71, pp.057301 (2005).

    """
    # Only apply to throats with an area
    ts = target.throats()[target[throat_area] > 0]
    P = target[throat_perimeter]
    A = target[throat_area]
    C = _np.ones(target.num_throats())
    C[ts] = P[ts]**2/A[ts]
    alpha = _np.ones_like(C)*8*_sp.pi
    if 'throat.offset_vertices' in target.props():
        verts = target['throat.offset_vertices']
        for i in ts:
            if ~_np.any(_np.isnan(verts[i])):
                if len(verts[i]) == 3:
                    # Triangular Correction
                    alpha[i] = C[i]*(25/17) + (40*_np.sqrt(3)/17)
                elif len(verts[i]) == 4:
                    # Rectangular Correction
                    alpha[i] = C[i]*(22/7) - (65/3)
                elif len(verts[i]) > 4:
                    # Approximate Elliptical Correction
                    alpha[i] = C[i]*(8/3) - (8*_sp.pi/3)
    # For a perfect circle alpha = 8*pi so normalize by this
    alpha /= 8*_sp.pi
    # Very small throats could have values less than one
    alpha[alpha < 1.0] = 1.0
    return alpha


def mason_morrow(target, throat_perimeter='throat.perimeter',
                 throat_area='throat.area'):
    r"""
    Mason and Morrow relate the capillary pressure to the shaped factor in a
    similar way to Mortensen but for triangles.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_perimeter : string
        The dictionary key of the array containing the throat perimeter values.

    throat_area : string
        The dictionary key of the array containing the throat area values.

    Returns
    -------
    value : NumPy ndarray
        Array containing throat shape factor values.

    References
    ----------
    Mason, G. and Morrow, N.R.. Capillary behavior of a perfectly wetting
    liquid in irregular triangular tubes. Journal of Colloid and Interface
    Science, 141(1), pp.262-274 (1991).

    """
    # Only apply to throats with an area
    ts = target.throats()[target[throat_area] <= 0]
    P = target[throat_perimeter]
    A = target[throat_area]
    value = A/(P**2)
    value[ts] = 1/(4*_sp.pi)
    return value


def jenkins_rao(target, throat_perimeter='throat.perimeter',
                throat_area='throat.area',
                throat_diameter='throat.indiameter'):
    r"""
    Jenkins and Rao relate the capillary pressure in an eliptical throat to
    the aspect ratio

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_perimeter : string
        The dictionary key of the array containing the throat perimeter values.

    throat_area : string
        The dictionary key of the array containing the throat area values.

    throat_diameter : string
        The dictionary key of the array containing the throat diameter values.

    Returns
    -------
    value : NumPy ndarray
        Array containing throat capillary pressure values.

    References
    ----------
    Jenkins, R.G. and Rao, M.B., The effect of elliptical pores on
    mercury porosimetry results. Powder technology, 38(2), pp.177-180. (1984)

    """
    P = target[throat_perimeter]
    A = target[throat_area]
    r = target[throat_diameter]/2
    # Normalized by value for perfect circle
    value = (P/A)/(2/r)
    return value



from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging, Docorator, GenericSettings
docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.get_sectionsf('FickianDiffusionSettings',
                      sections=['Parameters'])
@docstr.dedent
class FickianDiffusionSettings(GenericSettings):
    r"""

    Parameters
    ----------
    %(GenericTransportSettings.parameters)s
    quantity : str (default = 'pore.concentration')
        The name of the physical quantity to be calculated
    conductance : str (default = 'throat.diffusive_conductance')
        The name of the pore-scale transport conductance values. These are
        typically calculated by a model attached to a *Physics* object
        associated with the given *Phase*.

    Other Parameters
    ----------------

    **The following parameters pertain to the ReactiveTransport class**

    %(ReactiveTransportSettings.other_parameters)s

    ----

    **The following parameters pertain to the GenericTransport class**

    %(GenericTransportSettings.other_parameters)s

    """
    quantity = 'pore.concentration'
    conductance = 'throat.diffusive_conductance'


@docstr.dedent
class FickianDiffusion(ReactiveTransport):
    r"""
    A class to simulate binary diffusion with reactions

    Parameters
    ----------
    %(ReactiveTransport.parameters)s

    Notes
    -----
    Fickian diffusion in porous materials occurs in the void space, but
    becuase the diffusion is defined to pores it is impacted by the porosity
    and tortuosity of the network.  Thus the total diffusive flux through the
    network is reduced.  This class can be used to simualte diffusion-reaction
    in domains with arbitrarily complex boundary conditions, or it can be used
    to calculate the effective diffusivity of the network by applying
    controlled boundary conditions on opposing faces, calculate the diffusion
    rate, and inverting Fick's first law:

    .. math::

        D_{eff} = N_{A}*L/(A*\Delta C_{A})

    This class includes a method for calculating Deff automatically assuming
    appropriate boundary conditions were applied (``calc_eff_diffusivity``).
    The length and area of the domain should be supplied, but if they are
    not an attempt is made to calculate them.

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings._update_settings_and_docs(FickianDiffusionSettings())
        self.settings.update(settings)

    def setup(self, phase=None, quantity='', conductance='', **kwargs):
        r"""
        This method takes several arguments that are essential to running the
        algorithm and adds them to the settings.

        Parameters
        ----------
        %(FickianDiffusionSettings.parameters)s

        Notes
        -----
        Any additional arguments are added to the ``settings`` dictionary of
        the object.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        super().setup(**kwargs)

    def calc_effective_diffusivity(self, inlets=None, outlets=None,
                                   domain_area=None, domain_length=None):
        r"""
        This calculates the effective diffusivity in this linear transport
        algorithm.

        Parameters
        ----------
        inlets : array_like
            The pores where the inlet composition boundary conditions were
            applied.  If not given an attempt is made to infer them from the
            algorithm.

        outlets : array_like
            The pores where the outlet composition boundary conditions were
            applied.  If not given an attempt is made to infer them from the
            algorithm.

        domain_area : scalar, optional
            The area of the inlet (and outlet) boundary faces.  If not given
            then an attempt is made to estimate it, but it is usually
            underestimated.

        domain_length : scalar, optional
            The length of the domain between the inlet and outlet boundary
            faces.  If not given then an attempt is made to estimate it, but it
            is usually underestimated.

        Notes
        -----
        The area and length of the domain are found using the bounding box
        around the inlet and outlet pores which do not necessarily lie on the
        edge of the domain, resulting in underestimation of sizes.
        """
        return self._calc_eff_prop(inlets=inlets, outlets=outlets,
                                   domain_area=domain_area,
                                   domain_length=domain_length)
